# -*- coding: utf-8 -*-
# Copyright (C) 2006-2010 Søren Roug, European Environment Agency
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Contributor(s):
#
# Copyright (C) 2014 Georges Khaznadar <georgesk@debian.org>
#     migration to Python3, JavaDOC comments and automatic
#      build of documentation
#

__doc__="""Use OpenDocument to generate your documents."""

import zipfile, time, uuid, sys, mimetypes, copy, os.path

# to allow Python3 to access modules in the same path
sys.path.append(os.path.dirname(__file__))

# using BytesIO provides a cleaner interface than StringIO
# with both Python2 and Python3: the programmer must care to
# convert strings or unicode to bytes, which is valid for Python 2 and 3.
from io import StringIO, BytesIO

from odf.namespaces import *
import odf.manifest as manifest
import odf.meta as meta
from odf.office import *
import odf.element as element
from odf.attrconverters import make_NCName
from xml.sax.xmlreader import InputSource
from odf.odfmanifest import manifestlist
import codecs

if sys.version_info[0] == 3:
    unicode=str # unicode function does not exist

__version__= TOOLSVERSION

_XMLPROLOGUE = u"<?xml version='1.0' encoding='UTF-8'?>\n"

#####
# file permission as an integer value.
# The following syntax would be invalid for Python3:
# UNIXPERMS = 0100644 << 16L  # -rw-r--r--
#
# So it has been precomputed:
# 2175008768 is the same value as 0100644 << 16L == -rw-r--r--
####
UNIXPERMS = 2175008768

IS_FILENAME = 0
IS_IMAGE = 1
# We need at least Python 2.2
assert sys.version_info[0]>=2 and sys.version_info[1] >= 2

#sys.setrecursionlimit(100)
#The recursion limit is set conservative so mistakes like
# s=content() s.addElement(s) won't eat up too much processor time.

###############
# mime-types => file extensions
###############
odmimetypes = {
 u'application/vnd.oasis.opendocument.text':                  u'.odt',
 u'application/vnd.oasis.opendocument.text-template':         u'.ott',
 u'application/vnd.oasis.opendocument.graphics':              u'.odg',
 u'application/vnd.oasis.opendocument.graphics-template':     u'.otg',
 u'application/vnd.oasis.opendocument.presentation':          u'.odp',
 u'application/vnd.oasis.opendocument.presentation-template': u'.otp',
 u'application/vnd.oasis.opendocument.spreadsheet':           u'.ods',
 u'application/vnd.oasis.opendocument.spreadsheet-template':  u'.ots',
 u'application/vnd.oasis.opendocument.chart':                 u'.odc',
 u'application/vnd.oasis.opendocument.chart-template':        u'.otc',
 u'application/vnd.oasis.opendocument.image':                 u'.odi',
 u'application/vnd.oasis.opendocument.image-template':        u'.oti',
 u'application/vnd.oasis.opendocument.formula':               u'.odf',
 u'application/vnd.oasis.opendocument.formula-template':      u'.otf',
 u'application/vnd.oasis.opendocument.text-master':           u'.odm',
 u'application/vnd.oasis.opendocument.text-web':              u'.oth',
}

class OpaqueObject:
    """
    just a record to bear a filename, a mediatype and a bytes content
    """
    def __init__(self, filename, mediatype, content=None):
        """
        the constructor
        @param filename a unicode string
        @param mediatype a unicode string
        @param content a byte string or None
        """
        assert(type(filename)==type(u""))
        assert(type(mediatype)==type(u""))
        assert(type(content)==type(b"") or content == None)

        self.mediatype = mediatype
        self.filename = filename
        self.content = content

class OpenDocument:
    """
    A class to hold the content of an OpenDocument document
    Use the xml method to write the XML
    source to the screen or to a file.
    Example of use: d = OpenDocument(mimetype); fd.write(d.xml())
    """
    thumbnail = None

    def __init__(self, mimetype, add_generator=True):
        """
        the constructor
        @param mimetype a unicode string
        @param add_generator a boolean
        """
        assert(type(mimetype)==type(u""))
        assert(isinstance(add_generator,True.__class__))

        self.mimetype = mimetype
        self.childobjects = []
        self._extra = []
        self.folder = u"" # Always empty for toplevel documents
        self.topnode = Document(mimetype=self.mimetype)
        self.topnode.ownerDocument = self

        self.clear_caches()

        self.Pictures = {}
        self.meta = Meta()
        self.topnode.addElement(self.meta)
        if add_generator:
            self.meta.addElement(meta.Generator(text=TOOLSVERSION))
        self.scripts = Scripts()
        self.topnode.addElement(self.scripts)
        self.fontfacedecls = FontFaceDecls()
        self.topnode.addElement(self.fontfacedecls)
        self.settings = Settings()
        self.topnode.addElement(self.settings)
        self.styles = Styles()
        self.topnode.addElement(self.styles)
        self.automaticstyles = AutomaticStyles()
        self.topnode.addElement(self.automaticstyles)
        self.masterstyles = MasterStyles()
        self.topnode.addElement(self.masterstyles)
        self.body = Body()
        self.topnode.addElement(self.body)

    def rebuild_caches(self, node=None):
        if node is None: node = self.topnode
        self.build_caches(node)
        for e in node.childNodes:
            if e.nodeType == element.Node.ELEMENT_NODE:
                self.rebuild_caches(e)

    def clear_caches(self):
        """
        Clears internal caches
        """
        self.element_dict = {}
        self._styles_dict = {}
        self._styles_ooo_fix = {}

    def build_caches(self, elt):
        """
        Builds internal caches; called from element.py
        @param elt an element.Element instance
        """
        # assert(isinstance(elt, element.Element))
        # why do I need this more intricated assertion?
        # with Python3, the type of elt pops out as odf.element.Element
        # in one test ???
        import odf.element
        assert(isinstance(elt, element.Element) or isinstance(elt, odf.element.Element) )

        if elt.qname not in self.element_dict:
            self.element_dict[elt.qname] = []
        self.element_dict[elt.qname].append(elt)
        if elt.qname == (STYLENS, u'style'):
            self.__register_stylename(elt) # Add to style dictionary
        styleref = elt.getAttrNS(TEXTNS,u'style-name')
        if styleref is not None and styleref in self._styles_ooo_fix:
            elt.setAttrNS(TEXTNS,u'style-name', self._styles_ooo_fix[styleref])

    def remove_from_caches(self, elt):
        """
        Updates internal caches when an element has been removed
        @param elt an element.Element instance
        """
        # See remark in build_caches about the following assertion
        import odf.element
        assert(isinstance(elt, element.Element) or isinstance(elt, odf.element.Element))

        self.element_dict[elt.qname].remove(elt)
        for e in elt.childNodes:
            if e.nodeType == element.Node.ELEMENT_NODE:
                self.remove_from_caches(e)

        if elt.qname == (STYLENS, u'style'):
            del self._styles_dict[elt.getAttrNS(STYLENS, u'name')]

    def __register_stylename(self, elt):
        '''
        Register a style. But there are three style dictionaries:
        office:styles, office:automatic-styles and office:master-styles
        Chapter 14.
        @param elt an element.Element instance
        '''
        assert(isinstance(elt, element.Element))

        name = elt.getAttrNS(STYLENS, u'name')
        if name is None:
            return
        if elt.parentNode.qname in ((OFFICENS,u'styles'), (OFFICENS,u'automatic-styles')):
            if name in self._styles_dict:
                newname = u'M'+name # Rename style
                self._styles_ooo_fix[name] = newname
                # From here on all references to the old name will refer to the new one
                name = newname
                elt.setAttrNS(STYLENS, u'name', name)
            self._styles_dict[name] = elt

    def toXml(self, filename=u''):
        """
        converts the document to a valid Xml format.
        @param filename unicode string: the name of a file, defaults to
        an empty string.
        @return if filename is not empty, the XML code will be written into it
        and the method returns None; otherwise the method returns a StringIO
        containing valid XML.
        Then a ".getvalue()" should return a unicode string.
        """
        assert(type(filename)==type(u""))

        result=None
        xml=StringIO()
        if sys.version_info[0]==2:
            xml.write(_XMLPROLOGUE)
        else:
            xml.write(_XMLPROLOGUE)
        self.body.toXml(0, xml)
        if not filename:
            result=xml.getvalue()
        else:
            f=codecs.open(filename,'w', encoding='utf-8')
            f.write(xml.getvalue())
            f.close()
        return result

    def xml(self):
        """
        Generates the full document as an XML "file"
        @return a bytestream in UTF-8 encoding
        """
        self.__replaceGenerator()
        xml=StringIO()
        if sys.version_info[0]==2:
            xml.write(_XMLPROLOGUE)
        else:
            xml.write(_XMLPROLOGUE)
        self.topnode.toXml(0, xml)
        return xml.getvalue().encode("utf-8")


    def contentxml(self):
        """
        Generates the content.xml file
        @return a bytestream in UTF-8 encoding
        """
        xml=StringIO()
        xml.write(_XMLPROLOGUE)
        x = DocumentContent()
        x.write_open_tag(0, xml)
        if self.scripts.hasChildNodes():
            self.scripts.toXml(1, xml)
        if self.fontfacedecls.hasChildNodes():
            self.fontfacedecls.toXml(1, xml)
        a = AutomaticStyles()
        stylelist = self._used_auto_styles([self.styles, self.automaticstyles, self.body])
        if len(stylelist) > 0:
            a.write_open_tag(1, xml)
            for s in stylelist:
                s.toXml(2, xml)
            a.write_close_tag(1, xml)
        else:
            a.toXml(1, xml)
        self.body.toXml(1, xml)
        x.write_close_tag(0, xml)
        return xml.getvalue().encode("utf-8")

    def __manifestxml(self):
        """
        Generates the manifest.xml file;
        The self.manifest isn't avaible unless the document is being saved
        @return a unicode string
        """
        xml=StringIO()
        xml.write(_XMLPROLOGUE)
        self.manifest.toXml(0,xml)
        result=xml.getvalue()
        assert(type(result)==type(u""))
        return result

    def metaxml(self):
        """
        Generates the meta.xml file
        @return a unicode string
        """
        self.__replaceGenerator()
        x = DocumentMeta()
        x.addElement(self.meta)
        xml=StringIO()
        xml.write(_XMLPROLOGUE)
        x.toXml(0,xml)
        result=xml.getvalue()
        assert(type(result)==type(u""))
        return result

    def settingsxml(self):
        """
        Generates the settings.xml file
        @return a unicode string
        """
        x = DocumentSettings()
        x.addElement(self.settings)
        xml=StringIO()
        if sys.version_info[0]==2:
            xml.write(_XMLPROLOGUE)
        else:
            xml.write(_XMLPROLOGUE)
        x.toXml(0,xml)
        result=xml.getvalue()
        assert(type(result)==type(u""))
        return result

    def _parseoneelement(self, top, stylenamelist):
        """
        Finds references to style objects in master-styles
        and add the style name to the style list if not already there.
        Recursive
        @return the list of style names as unicode strings
        """
        for e in top.childNodes:
            if e.nodeType == element.Node.ELEMENT_NODE:
                for styleref in (
                        (CHARTNS,u'style-name'),
                        (DRAWNS,u'style-name'),
                        (DRAWNS,u'text-style-name'),
                        (PRESENTATIONNS,u'style-name'),
                        (STYLENS,u'data-style-name'),
                        (STYLENS,u'list-style-name'),
                        (STYLENS,u'page-layout-name'),
                        (STYLENS,u'style-name'),
                        (TABLENS,u'default-cell-style-name'),
                        (TABLENS,u'style-name'),
                        (TEXTNS,u'style-name') ):
                    if e.getAttrNS(styleref[0],styleref[1]):
                        stylename = e.getAttrNS(styleref[0],styleref[1])
                        if stylename not in stylenamelist:
                            # due to the polymorphism of e.getAttrNS(),
                            # a unicode type is enforced for elements
                            stylenamelist.append(unicode(stylename))
                stylenamelist = self._parseoneelement(e, stylenamelist)
        return stylenamelist

    def _used_auto_styles(self, segments):
        """
        Loop through the masterstyles elements, and find the automatic
        styles that are used. These will be added to the automatic-styles
        element in styles.xml
        @return a list of element.Element instances
        """
        stylenamelist = []
        for top in segments:
            stylenamelist = self._parseoneelement(top, stylenamelist)
        stylelist = []
        for e in self.automaticstyles.childNodes:
            if isinstance(e, element.Element) and e.getAttrNS(STYLENS,u'name') in stylenamelist:
                stylelist.append(e)

        # check the type of the returned data
        ok=True
        for e in stylelist: ok = ok and isinstance(e, element.Element)
        assert(ok)

        return stylelist

    def stylesxml(self):
        """
        Generates the styles.xml file
        @return valid XML code as a unicode string
        """
        xml=StringIO()
        xml.write(_XMLPROLOGUE)
        x = DocumentStyles()
        x.write_open_tag(0, xml)
        if self.fontfacedecls.hasChildNodes():
            self.fontfacedecls.toXml(1, xml)
        self.styles.toXml(1, xml)
        a = AutomaticStyles()
        a.write_open_tag(1, xml)
        for s in self._used_auto_styles([self.masterstyles]):
            s.toXml(2, xml)
        a.write_close_tag(1, xml)
        if self.masterstyles.hasChildNodes():
            self.masterstyles.toXml(1, xml)
        x.write_close_tag(0, xml)
        result = xml.getvalue()

        assert(type(result)==type(u""))

        return result

    def addPicture(self, filename, mediatype=None, content=None):
        """
        Add a picture
        It uses the same convention as OOo, in that it saves the picture in
        the zipfile in the subdirectory 'Pictures'
        If passed a file ptr, mediatype must be set
        @param filename unicode string: name of a file for Pictures
        @param mediatype unicode string: name of a media, None by default
        @param content bytes: content of media, None by default
        @return a unicode string: the file name of the media, eventually
        created on the fly
        """
        if content is None:
            if mediatype is None:
                mediatype, encoding = mimetypes.guess_type(filename)
            if mediatype is None:
                mediatype = u''
                try: ext = filename[filename.rindex(u'.'):]
                except: ext=u''
            else:
                ext = mimetypes.guess_extension(mediatype)
            manifestfn = u"Pictures/%s%s" % (uuid.uuid4().hex.upper(), ext)
            self.Pictures[manifestfn] = (IS_FILENAME, filename, mediatype)
            content=b""  # this value is only use by the assert further
            filename=u"" # this value is only use by the assert further
        else:
            manifestfn = filename
            self.Pictures[manifestfn] = (IS_IMAGE, content, mediatype)

        assert(type(filename)==type(u""))
        assert(type(content) == type(b""))

        return manifestfn

    def addPictureFromFile(self, filename, mediatype=None):
        """
        Add a picture
        It uses the same convention as OOo, in that it saves the picture in
        the zipfile in the subdirectory 'Pictures'.
        If mediatype is not given, it will be guessed from the filename
        extension.
        @param filesname unicode string: name of an image file
        @param mediatype unicode string: type of media, dfaults to None
        @return a unicode string, the name of the created file
        """
        if mediatype is None:
            mediatype, encoding = mimetypes.guess_type(filename)
        if mediatype is None:
            mediatype = u''
            try: ext = filename[filename.rindex(u'.'):]
            except ValueError: ext=u''
        else:
            ext = mimetypes.guess_extension(mediatype)
        manifestfn = u"Pictures/%s%s" % (uuid.uuid4().hex.upper(), ext)
        self.Pictures[manifestfn] = (IS_FILENAME, filename, mediatype)

        assert(type(filename)==type(u""))
        assert(type(mediatype)==type(u""))

        return manifestfn

    def addPictureFromString(self, content, mediatype):
        """
        Add a picture from contents given as a Byte string.
        It uses the same convention as OOo, in that it saves the picture in
        the zipfile in the subdirectory 'Pictures'. The content variable
        is a string that contains the binary image data. The mediatype
        indicates the image format.
        @param content bytes: content of media
        @param mediatype unicode string: name of a media
        @return a unicode string, the name of the created file
        """
        assert(type(content)==type(b""))
        assert(type(mediatype)==type(u""))

        ext = mimetypes.guess_extension(mediatype)
        manifestfn = u"Pictures/%s%s" % (uuid.uuid4().hex.upper(), ext)
        self.Pictures[manifestfn] = (IS_IMAGE, content, mediatype)
        return manifestfn

    def addThumbnail(self, filecontent=None):
        """
        Add a fixed thumbnail
        The thumbnail in the library is big, so this is pretty useless.
        @param filecontent bytes: the content of a file; defaults to None
        """
        assert(type(filecontent)==type(b""))

        if filecontent is None:
            import thumbnail
            self.thumbnail = thumbnail.thumbnail()
        else:
            self.thumbnail = filecontent

    def addObject(self, document, objectname=None):
        """
        Adds an object (subdocument). The object must be an OpenDocument class
        @param document OpenDocument instance
        @param objectname unicode string: the name of an object to add
        @return a unicode string: the folder name in the zipfile the object is
        stored in.
        """
        assert(isinstance(document, OpenDocument))
        assert(type(objectname)==type(u"") or objectname == None)

        self.childobjects.append(document)
        if objectname is None:
            document.folder = u"%s/Object %d" % (self.folder, len(self.childobjects))
        else:
            document.folder = objectname
        return u".%s" % document.folder

    def _savePictures(self, anObject, folder):
        """
        saves pictures contained in an object
        @param anObject instance of OpenDocument containing pictures
        @param folder unicode string: place to save pictures
        """
        assert(isinstance(anObject, OpenDocument))
        assert(type(folder)==type(u""))

        hasPictures = False
        for arcname, picturerec in anObject.Pictures.items():
            what_it_is, fileobj, mediatype = picturerec
            self.manifest.addElement(manifest.FileEntry(fullpath=u"%s%s" % ( folder ,arcname), mediatype=mediatype))
            hasPictures = True
            if what_it_is == IS_FILENAME:
                self._z.write(fileobj, folder + arcname, zipfile.ZIP_STORED)
            else:
                zi = zipfile.ZipInfo(str(arcname), self._now)
                zi.compress_type = zipfile.ZIP_STORED
                zi.external_attr = UNIXPERMS
                self._z.writestr(zi, fileobj)
        # According to section 17.7.3 in ODF 1.1, the pictures folder should not have a manifest entry
#       if hasPictures:
#           self.manifest.addElement(manifest.FileEntry(fullpath="%sPictures/" % folder, mediatype=""))
        # Look in subobjects
        subobjectnum = 1
        for subobject in anObject.childobjects:
            self._savePictures(subobject, u'%sObject %d/' % (folder, subobjectnum))
            subobjectnum += 1

    def __replaceGenerator(self):
        """
        Removes a previous 'generator' stance and declares TOOLSVERSION
        as the new generator.
        Section 3.1.1: The application MUST NOT export the original identifier
        belonging to the application that created the document.
        """
        for m in self.meta.childNodes[:]:
            if hasattr(m,'qname') and m.qname == (METANS, u'generator'):
                self.meta.removeChild(m)
        self.meta.addElement(meta.Generator(text=TOOLSVERSION))

    def save(self, outputfile, addsuffix=False):
        """
        Save the document under the filename.
        If the filename is '-' then save to stdout
        @param outputfile unicode string: the special name '-' is for stdout;
        as an alternative, it can be an io.ByteIO instance which contains
        the ZIP content.
        @param addsuffix boolean: whether to add a suffix or not; defaults to False
        """

        if outputfile == u'-':
            outputfp = zipfile.ZipFile(sys.stdout,"w")
        else:
            if addsuffix:
                outputfile = outputfile + odmimetypes.get(self.mimetype,u'.xxx')
            outputfp = zipfile.ZipFile(outputfile, "w")
        self.__zipwrite(outputfp)
        outputfp.close()

    def write(self, outputfp):
        """
        User API to write the ODF file to an open file descriptor
        Writes the ZIP format
        @param outputfp open file descriptor
        """
        zipoutputfp = zipfile.ZipFile(outputfp,"w")
        self.__zipwrite(zipoutputfp)

    def __zipwrite(self, outputfp):
        """
        Write the document to an open file pointer
        This is where the real work is done
        @param outputfp instance of zipfile.ZipFile
        """
        assert(isinstance(outputfp, zipfile.ZipFile))

        self._z = outputfp
        self._now = time.localtime()[:6]
        self.manifest = manifest.Manifest()

        # Write mimetype
        zi = zipfile.ZipInfo('mimetype', self._now)
        zi.compress_type = zipfile.ZIP_STORED
        zi.external_attr = UNIXPERMS
        self._z.writestr(zi, self.mimetype.encode("utf-8"))

        self._saveXmlObjects(self,u"")

        # Write pictures
        self._savePictures(self,u"")

        # Write the thumbnail
        if self.thumbnail is not None:
            self.manifest.addElement(manifest.FileEntry(fullpath=u"Thumbnails/", mediatype=u''))
            self.manifest.addElement(manifest.FileEntry(fullpath=u"Thumbnails/thumbnail.png", mediatype=u''))
            zi = zipfile.ZipInfo(u"Thumbnails/thumbnail.png", self._now)
            zi.compress_type = zipfile.ZIP_DEFLATED
            zi.external_attr = UNIXPERMS
            self._z.writestr(zi, self.thumbnail)

        # Write any extra files
        for op in self._extra:
            if op.filename == u"META-INF/documentsignatures.xml": continue # Don't save signatures
            self.manifest.addElement(manifest.FileEntry(fullpath=op.filename, mediatype=op.mediatype))
            if sys.version_info[0]==3:
                zi = zipfile.ZipInfo(op.filename, self._now)
            else:
                zi = zipfile.ZipInfo(op.filename.encode('utf-8'), self._now)
            zi.compress_type = zipfile.ZIP_DEFLATED
            zi.external_attr = UNIXPERMS
            if op.content is not None:
                self._z.writestr(zi, op.content)
        # Write manifest
        zi = zipfile.ZipInfo(u"META-INF/manifest.xml", self._now)
        zi.compress_type = zipfile.ZIP_DEFLATED
        zi.external_attr = UNIXPERMS
        self._z.writestr(zi, self.__manifestxml() )
        del self._z
        del self._now
        del self.manifest


    def _saveXmlObjects(self, anObject, folder):
        """
        save xml objects of an opendocument to some folder
        @param anObject instance of OpenDocument
        @param folder unicode string place to save xml objects
        """
        assert(isinstance(anObject, OpenDocument))
        assert(type(folder)==type(u""))

        if self == anObject:
            self.manifest.addElement(manifest.FileEntry(fullpath=u"/", mediatype=anObject.mimetype))
        else:
            self.manifest.addElement(manifest.FileEntry(fullpath=folder, mediatype=anObject.mimetype))
        # Write styles
        self.manifest.addElement(manifest.FileEntry(fullpath=u"%sstyles.xml" % folder, mediatype=u"text/xml"))
        zi = zipfile.ZipInfo(u"%sstyles.xml" % folder, self._now)
        zi.compress_type = zipfile.ZIP_DEFLATED
        zi.external_attr = UNIXPERMS
        self._z.writestr(zi, anObject.stylesxml().encode("utf-8") )

        # Write content
        self.manifest.addElement(manifest.FileEntry(fullpath=u"%scontent.xml" % folder, mediatype=u"text/xml"))
        zi = zipfile.ZipInfo(u"%scontent.xml" % folder, self._now)
        zi.compress_type = zipfile.ZIP_DEFLATED
        zi.external_attr = UNIXPERMS
        self._z.writestr(zi, anObject.contentxml() )

        # Write settings
        if anObject.settings.hasChildNodes():
            self.manifest.addElement(manifest.FileEntry(fullpath=u"%ssettings.xml" % folder, mediatype=u"text/xml"))
            zi = zipfile.ZipInfo(u"%ssettings.xml" % folder, self._now)
            zi.compress_type = zipfile.ZIP_DEFLATED
            zi.external_attr = UNIXPERMS
            self._z.writestr(zi, anObject.settingsxml().encode("utf-8") )

        # Write meta
        if self == anObject:
            self.manifest.addElement(manifest.FileEntry(fullpath=u"meta.xml", mediatype=u"text/xml"))
            zi = zipfile.ZipInfo(u"meta.xml", self._now)
            zi.compress_type = zipfile.ZIP_DEFLATED
            zi.external_attr = UNIXPERMS
            self._z.writestr(zi, anObject.metaxml().encode("utf-8") )

        # Write subobjects
        subobjectnum = 1
        for subobject in anObject.childobjects:
            self._saveXmlObjects(subobject, u'%sObject %d/' % (folder, subobjectnum))
            subobjectnum += 1

# Document's DOM methods
    def createElement(self, elt):
        """
        Inconvenient interface to create an element, but follows XML-DOM.
        Does not allow attributes as argument, therefore can't check grammar.
        @param elt element.Element instance
        @return an element.Element instance whose grammar is not checked
        """
        assert(isinstance(elt, element.Element))

        # this old code is ambiguous: is 'element' the module or is it the
        # local variable? To disambiguate this, the local variable has been
        # renamed to 'elt'
        #return element(check_grammar=False)
        return elt(check_grammar=False)

    def createTextNode(self, data):
        """
        Method to create a text node
        @param data unicode string to include in the Text element
        @return an instance of element.Text
        """
        assert(type(data)==type(u""))

        return element.Text(data)

    def createCDATASection(self, data):
        """
        Method to create a CDATA section
        @param data unicode string to include in the CDATA element
        @return an instance of element.CDATASection
        """
        assert(type(data)==type(u""))

        return element.CDATASection(cdata)

    def getMediaType(self):
        """
        Returns the media type
        @result a unicode string
        """
        assert (type(self.mimetype)==type(u""))

        return self.mimetype

    def getStyleByName(self, name):
        """
        Finds a style object based on the name
        @param name unicode string the name of style to search
        @return a syle as an element.Element instance
        """
        assert(type(name)==type(u""))

        ncname = make_NCName(name)
        if self._styles_dict == {}:
            self.rebuild_caches()
        result=self._styles_dict.get(ncname, None)

        assert(isinstance(result, element.Element))
        return result

    def getElementsByType(self, elt):
        """
        Gets elements based on the type, which is function from
        text.py, draw.py etc.
        @param elt instance of a function which returns an element.Element
        @return a list of istances of element.Element
        """
        import types
        assert(isinstance (elt, types.FunctionType))

        obj = elt(check_grammar=False)
        assert (isinstance(obj, element.Element))

        if self.element_dict == {}:
            self.rebuild_caches()

        # This previous code was ambiguous
        # was "element" the module name or the local variable?
        # the local variable is renamed to "elt" to disambiguate the code
        #return self.element_dict.get(obj.qname, [])

        result=self.element_dict.get(obj.qname, [])

        ok=True
        for e in result: ok = ok and isinstance(e, element.Element)
        assert(ok)

        return result

# Convenience functions
def OpenDocumentChart():
    """
    Creates a chart document
    @return an OpenDocument instance with chart mimetype
    """
    doc = OpenDocument(u'application/vnd.oasis.opendocument.chart')
    doc.chart = Chart()
    doc.body.addElement(doc.chart)
    return doc

def OpenDocumentDrawing():
    """
    Creates a drawing document
    @return an OpenDocument instance with drawing mimetype
    """
    doc = OpenDocument(u'application/vnd.oasis.opendocument.graphics')
    doc.drawing = Drawing()
    doc.body.addElement(doc.drawing)
    return doc

def OpenDocumentImage():
    """
    Creates an image document
    @return an OpenDocument instance with image mimetype
    """
    doc = OpenDocument(u'application/vnd.oasis.opendocument.image')
    doc.image = Image()
    doc.body.addElement(doc.image)
    return doc

def OpenDocumentPresentation():
    """
    Creates a presentation document
    @return an OpenDocument instance with presentation mimetype
    """
    doc = OpenDocument(u'application/vnd.oasis.opendocument.presentation')
    doc.presentation = Presentation()
    doc.body.addElement(doc.presentation)
    return doc

def OpenDocumentSpreadsheet():
    """
    Creates a spreadsheet document
    @return an OpenDocument instance with spreadsheet mimetype
    """
    doc = OpenDocument(u'application/vnd.oasis.opendocument.spreadsheet')
    doc.spreadsheet = Spreadsheet()
    doc.body.addElement(doc.spreadsheet)
    return doc

def OpenDocumentText():
    """
    Creates a text document
    @return an OpenDocument instance with text mimetype
    """
    doc = OpenDocument(u'application/vnd.oasis.opendocument.text')
    doc.text = Text()
    doc.body.addElement(doc.text)
    return doc

def OpenDocumentTextMaster():
    """
    Creates a text master document
    @return an OpenDocument instance with master mimetype
    """
    doc = OpenDocument(u'application/vnd.oasis.opendocument.text-master')
    doc.text = Text()
    doc.body.addElement(doc.text)
    return doc

def __loadxmlparts(z, manifest, doc, objectpath):
    """
    Parses a document from its zipfile
    @param z an instance of zipfile.ZipFile
    @param manifest Manifest data structured in a dictionary
    @param doc instance of OpenDocument to feed in
    @param objectpath unicode string: path to an object
    """
    assert(isinstance(z, zipfile.ZipFile))
    assert(type(manifest)==type(dict()))
    assert(isinstance(doc, OpenDocument))
    assert(type(objectpath)==type(u""))

    from odf.load import LoadParser
    from defusedxml.sax import make_parser
    from xml.sax import handler

    for xmlfile in (objectpath+u'settings.xml', objectpath+u'meta.xml', objectpath+u'content.xml', objectpath+u'styles.xml'):
        if xmlfile not in manifest:
            continue
        ##########################################################
        # this one is added to debug the bad behavior with Python2
        # which raises exceptions of type SAXParseException
        from xml.sax._exceptions import SAXParseException
        ##########################################################
        try:
            xmlpart = z.read(xmlfile).decode("utf-8")
            doc._parsing = xmlfile

            parser = make_parser()
            parser.setFeature(handler.feature_namespaces, 1)
            parser.setFeature(handler.feature_external_ges, 0)
            parser.setContentHandler(LoadParser(doc))
            parser.setErrorHandler(handler.ErrorHandler())

            inpsrc = InputSource()
            #################
            # There may be a SAXParseException triggered because of
            # a missing xmlns prefix like meta, config, etc.
            # So i add such declarations when needed (GK, 2014/10/21).
            # Is there any option to prevent xmlns checks by SAX?
            xmlpart=__fixXmlPart(xmlpart)

            inpsrc.setByteStream(BytesIO(xmlpart.encode("utf-8")))
            parser.parse(inpsrc)
            del doc._parsing
        except KeyError as v: pass
        except SAXParseException:
            print (u"====== SAX FAILED TO PARSE ==========\n", xmlpart)

def __fixXmlPart(xmlpart):
    """
    fixes an xml code when it does not contain a set of requested
    "xmlns:whatever" declarations.
    added by G.K. on 2014/10/21
    @param xmlpart unicode string: some XML code
    @return fixed XML code
    """
    result=xmlpart
    requestedPrefixes = (u'meta', u'config', u'dc', u'style',
                         u'svg', u'fo',u'draw', u'table',u'form')
    for prefix in requestedPrefixes:
        if u' xmlns:{prefix}'.format(prefix=prefix) not in xmlpart:
            ###########################################
            # fixed a bug triggered by math elements
            # Notice: math elements are creectly exported to XHTML
            #         and best viewed with MathJax javascript.
            # 2016-02-19 G.K.
            ###########################################
            try:
                pos=result.index(u" xmlns:")
                toInsert=u' xmlns:{prefix}="urn:oasis:names:tc:opendocument:xmlns:{prefix}:1.0"'.format(prefix=prefix)
                result=result[:pos]+toInsert+result[pos:]
            except:
                pass
    return result


def __detectmimetype(zipfd, odffile):
    """
    detects the mime-type of an ODF file
    @param zipfd an open zipfile.ZipFile instance
    @param odffile this parameter is not used
    @return a mime-type as a unicode string
    """
    assert(isinstance(zipfd, zipfile.ZipFile))

    try:
        mimetype = zipfd.read('mimetype').decode("utf-8")
        return mimetype
    except:
        pass
    # Fall-through to next mechanism
    manifestpart = zipfd.read('META-INF/manifest.xml')
    manifest =  manifestlist(manifestpart)
    for mentry,mvalue in manifest.items():
        if mentry == "/":
            assert(type(mvalue['media-type'])==type(u""))
            return mvalue['media-type']

    # Fall-through to last mechanism
    return u'application/vnd.oasis.opendocument.text'

def load(odffile):
    """
    Load an ODF file into memory
    @param odffile unicode string: name of a file, or as an alternative,
    an open readable stream
    @return a reference to the structure (an OpenDocument instance)
    """
    z = zipfile.ZipFile(odffile)
    mimetype = __detectmimetype(z, odffile)
    doc = OpenDocument(mimetype, add_generator=False)

    # Look in the manifest file to see if which of the four files there are
    manifestpart = z.read('META-INF/manifest.xml')
    manifest =  manifestlist(manifestpart)
    __loadxmlparts(z, manifest, doc, u'')
    for mentry,mvalue in manifest.items():
        if mentry[:9] == u"Pictures/" and len(mentry) > 9:
            doc.addPicture(mvalue['full-path'], mvalue['media-type'], z.read(mentry))
        elif mentry == u"Thumbnails/thumbnail.png":
            doc.addThumbnail(z.read(mentry))
        elif mentry in (u'settings.xml', u'meta.xml', u'content.xml', u'styles.xml'):
            pass
        # Load subobjects into structure
        elif mentry[:7] == u"Object " and len(mentry) < 11 and mentry[-1] == u"/":
            subdoc = OpenDocument(mvalue['media-type'], add_generator=False)
            doc.addObject(subdoc, u"/" + mentry[:-1])
            __loadxmlparts(z, manifest, subdoc, mentry)
        elif mentry[:7] == u"Object ":
            pass # Don't load subobjects as opaque objects
        else:
            if mvalue['full-path'][-1] == u'/':
                doc._extra.append(OpaqueObject(mvalue['full-path'], mvalue['media-type'], None))
            else:
                doc._extra.append(OpaqueObject(mvalue['full-path'], mvalue['media-type'], z.read(mentry)))
            # Add the SUN junk here to the struct somewhere
            # It is cached data, so it can be out-of-date
    z.close()
    b = doc.getElementsByType(Body)
    if mimetype[:39] == u'application/vnd.oasis.opendocument.text':
        doc.text = b[0].firstChild
    elif mimetype[:43] == u'application/vnd.oasis.opendocument.graphics':
        doc.graphics = b[0].firstChild
    elif mimetype[:47] == u'application/vnd.oasis.opendocument.presentation':
        doc.presentation = b[0].firstChild
    elif mimetype[:46] == u'application/vnd.oasis.opendocument.spreadsheet':
        doc.spreadsheet = b[0].firstChild
    elif mimetype[:40] == u'application/vnd.oasis.opendocument.chart':
        doc.chart = b[0].firstChild
    elif mimetype[:40] == u'application/vnd.oasis.opendocument.image':
        doc.image = b[0].firstChild
    elif mimetype[:42] == u'application/vnd.oasis.opendocument.formula':
        doc.formula = b[0].firstChild

    return doc

# vim: set expandtab sw=4 :



# -*- coding: utf-8 -*-
# Copyright (C) 2006-2007 Søren Roug, European Environment Agency
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s):
#

from odf.namespaces import SVGNS
from odf.element import Element
from odf.draw import DrawElement

# Autogenerated
def DefinitionSrc(**args):
    args.setdefault('type', 'simple')
    return Element(qname = (SVGNS,'definition-src'), **args)

def Desc(**args):
    return Element(qname = (SVGNS,'desc'), **args)

def FontFaceFormat(**args):
    return Element(qname = (SVGNS,'font-face-format'), **args)

def FontFaceName(**args):
    return Element(qname = (SVGNS,'font-face-name'), **args)

def FontFaceSrc(**args):
    return Element(qname = (SVGNS,'font-face-src'), **args)

def FontFaceUri(**args):
    args.setdefault('type', 'simple')
    return Element(qname = (SVGNS,'font-face-uri'), **args)

def Lineargradient(**args):
    return DrawElement(qname = (SVGNS,'linearGradient'), **args)

def Radialgradient(**args):
    return DrawElement(qname = (SVGNS,'radialGradient'), **args)

def Stop(**args):
    return Element(qname = (SVGNS,'stop'), **args)

def Title(**args):
    return Element(qname = (SVGNS,'title'), **args)



# -*- coding: utf-8 -*-
# Copyright (C) 2006-2013 Søren Roug, European Environment Agency
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s):
#

from odf.namespaces import OFFICENS
from odf.element import Element
from odf.draw import StyleRefElement

# Autogenerated
def Annotation(**args):
    return StyleRefElement(qname = (OFFICENS,'annotation'), **args)

def AnnotationEnd(**args):
    return StyleRefElement(qname = (OFFICENS,'annotation-end'), **args)

def AutomaticStyles(**args):
    return Element(qname = (OFFICENS, 'automatic-styles'), **args)

def BinaryData(**args):
    return Element(qname = (OFFICENS,'binary-data'), **args)

def Body(**args):
    return Element(qname = (OFFICENS, 'body'), **args)

def ChangeInfo(**args):
    return Element(qname = (OFFICENS,'change-info'), **args)

def Chart(**args):
    return Element(qname = (OFFICENS,'chart'), **args)

def DdeSource(**args):
    return Element(qname = (OFFICENS,'dde-source'), **args)

def Document(version="1.2", **args):
    return Element(qname = (OFFICENS,'document'), version=version, **args)

def DocumentContent(version="1.2", **args):
    return Element(qname = (OFFICENS, 'document-content'), version=version, **args)

def DocumentMeta(version="1.2", **args):
    return Element(qname = (OFFICENS, 'document-meta'), version=version, **args)

def DocumentSettings(version="1.2", **args):
    return Element(qname = (OFFICENS, 'document-settings'), version=version, **args)

def DocumentStyles(version="1.2", **args):
    return Element(qname = (OFFICENS, 'document-styles'), version=version, **args)

def Drawing(**args):
    return Element(qname = (OFFICENS,'drawing'), **args)

def EventListeners(**args):
    return Element(qname = (OFFICENS,'event-listeners'), **args)

def FontFaceDecls(**args):
    return Element(qname = (OFFICENS, 'font-face-decls'), **args)

def Forms(**args):
    return Element(qname = (OFFICENS,'forms'), **args)

def Image(**args):
    return Element(qname = (OFFICENS,'image'), **args)

def MasterStyles(**args):
    return Element(qname = (OFFICENS, 'master-styles'), **args)

def Meta(**args):
    return Element(qname = (OFFICENS, 'meta'), **args)

def Presentation(**args):
    return Element(qname = (OFFICENS,'presentation'), **args)

def Script(**args):
    return Element(qname = (OFFICENS, 'script'), **args)

def Scripts(**args):
    return Element(qname = (OFFICENS, 'scripts'), **args)

def Settings(**args):
    return Element(qname = (OFFICENS, 'settings'), **args)

def Spreadsheet(**args):
    return Element(qname = (OFFICENS, 'spreadsheet'), **args)

def Styles(**args):
    return Element(qname = (OFFICENS, 'styles'), **args)

def Text(**args):
    return Element(qname = (OFFICENS, 'text'), **args)

# Autogenerated end



# -*- coding: utf-8 -*-
# Copyright (C) 2006-2007 Søren Roug, European Environment Agency
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s):
#

from odf.namespaces import PRESENTATIONNS
from odf.element import Element

# ODF 1.0 section 9.6 and 9.7
# Autogenerated
def AnimationGroup(**args):
    return Element(qname = (PRESENTATIONNS,'animation-group'), **args)

def Animations(**args):
    return Element(qname = (PRESENTATIONNS,'animations'), **args)

def DateTime(**args):
    return Element(qname = (PRESENTATIONNS,'date-time'), **args)

def DateTimeDecl(**args):
    return Element(qname = (PRESENTATIONNS,'date-time-decl'), **args)

def Dim(**args):
    return Element(qname = (PRESENTATIONNS,'dim'), **args)

def EventListener(**args):
    return Element(qname = (PRESENTATIONNS,'event-listener'), **args)

def Footer(**args):
    return Element(qname = (PRESENTATIONNS,'footer'), **args)

def FooterDecl(**args):
    return Element(qname = (PRESENTATIONNS,'footer-decl'), **args)

def Header(**args):
    return Element(qname = (PRESENTATIONNS,'header'), **args)

def HeaderDecl(**args):
    return Element(qname = (PRESENTATIONNS,'header-decl'), **args)

def HideShape(**args):
    return Element(qname = (PRESENTATIONNS,'hide-shape'), **args)

def HideText(**args):
    return Element(qname = (PRESENTATIONNS,'hide-text'), **args)

def Notes(**args):
    return Element(qname = (PRESENTATIONNS,'notes'), **args)

def Placeholder(**args):
    return Element(qname = (PRESENTATIONNS,'placeholder'), **args)

def Play(**args):
    return Element(qname = (PRESENTATIONNS,'play'), **args)

def Settings(**args):
    return Element(qname = (PRESENTATIONNS,'settings'), **args)

def Show(**args):
    return Element(qname = (PRESENTATIONNS,'show'), **args)

def ShowShape(**args):
    return Element(qname = (PRESENTATIONNS,'show-shape'), **args)

def ShowText(**args):
    return Element(qname = (PRESENTATIONNS,'show-text'), **args)

def Sound(**args):
    args.setdefault('type', 'simple')
    return Element(qname = (PRESENTATIONNS,'sound'), **args)




# -*- coding: utf-8 -*-
# Copyright (C) 2006-2007 Søren Roug, European Environment Agency
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s):
#

from odf.namespaces import DCNS
from odf.element import Element

# Autogenerated
def Creator(**args):
    return Element(qname = (DCNS,'creator'), **args)

def Date(**args):
    return Element(qname = (DCNS,'date'), **args)

def Description(**args):
    return Element(qname = (DCNS,'description'), **args)

def Language(**args):
    return Element(qname = (DCNS,'language'), **args)

def Subject(**args):
    return Element(qname = (DCNS,'subject'), **args)

def Title(**args):
    return Element(qname = (DCNS,'title'), **args)

# The following complete the Dublin Core elements, but there is no
# guarantee a compliant implementation of OpenDocument will preserve
# these elements

#def Contributor(**args):
#    return Element(qname = (DCNS,'contributor'), **args)

#def Coverage(**args):
#    return Element(qname = (DCNS,'coverage'), **args)

#def Format(**args):
#    return Element(qname = (DCNS,'format'), **args)

#def Identifier(**args):
#    return Element(qname = (DCNS,'identifier'), **args)

#def Publisher(**args):
#    return Element(qname = (DCNS,'publisher'), **args)

#def Relation(**args):
#    return Element(qname = (DCNS,'relation'), **args)

#def Rights(**args):
#    return Element(qname = (DCNS,'rights'), **args)

#def Source(**args):
#    return Element(qname = (DCNS,'source'), **args)

#def Type(**args):
#    return Element(qname = (DCNS,'type'), **args)



# -*- coding: utf-8 -*-
# Copyright (C) 2006-2013 Søren Roug, European Environment Agency
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s):
#
__version__ = "1.4.2"

TOOLSVERSION = u"ODFPY/" + __version__

ANIMNS         = u"urn:oasis:names:tc:opendocument:xmlns:animation:1.0"
CHARTNS        = u"urn:oasis:names:tc:opendocument:xmlns:chart:1.0"
CHARTOOONS     = u"http://openoffice.org/2010/chart"
CONFIGNS       = u"urn:oasis:names:tc:opendocument:xmlns:config:1.0"
CSS3TNS        = u"http://www.w3.org/TR/css3-text/"
#DBNS           = u"http://openoffice.org/2004/database"
DBNS           = u"urn:oasis:names:tc:opendocument:xmlns:database:1.0"
DCNS           = u"http://purl.org/dc/elements/1.1/"
DOMNS          = u"http://www.w3.org/2001/xml-events"
DR3DNS         = u"urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0"
DRAWNS         = u"urn:oasis:names:tc:opendocument:xmlns:drawing:1.0"
FIELDNS        = u"urn:openoffice:names:experimental:ooo-ms-interop:xmlns:field:1.0"
FONS           = u"urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0"
FORMNS         = u"urn:oasis:names:tc:opendocument:xmlns:form:1.0"
FORMXNS        = u"urn:openoffice:names:experimental:ooxml-odf-interop:xmlns:form:1.0"
GRDDLNS        = u"http://www.w3.org/2003/g/data-view#"
KOFFICENS      = u"http://www.koffice.org/2005/"
LOEXTNS        = u"urn:org:documentfoundation:names:experimental:office:xmlns:loext:1.0"
MANIFESTNS     = u"urn:oasis:names:tc:opendocument:xmlns:manifest:1.0"
MATHNS         = u"http://www.w3.org/1998/Math/MathML"
METANS         = u"urn:oasis:names:tc:opendocument:xmlns:meta:1.0"
NUMBERNS       = u"urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0"
OFFICENS       = u"urn:oasis:names:tc:opendocument:xmlns:office:1.0"
OFNS           = u"urn:oasis:names:tc:opendocument:xmlns:of:1.2"
OOOCNS         = u"http://openoffice.org/2004/calc"
OOONS          = u"http://openoffice.org/2004/office"
OOOWNS         = u"http://openoffice.org/2004/writer"
PRESENTATIONNS = u"urn:oasis:names:tc:opendocument:xmlns:presentation:1.0"
RDFANS         = u"http://docs.oasis-open.org/opendocument/meta/rdfa#"
RPTNS          = u"http://openoffice.org/2005/report"
SCRIPTNS       = u"urn:oasis:names:tc:opendocument:xmlns:script:1.0"
SMILNS         = u"urn:oasis:names:tc:opendocument:xmlns:smil-compatible:1.0"
STYLENS        = u"urn:oasis:names:tc:opendocument:xmlns:style:1.0"
SVGNS          = u"urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0"
TABLENS        = u"urn:oasis:names:tc:opendocument:xmlns:table:1.0"
TABLEOOONS     = u"http://openoffice.org/2009/table"
TEXTNS         = u"urn:oasis:names:tc:opendocument:xmlns:text:1.0"
XFORMSNS       = u"http://www.w3.org/2002/xforms"
XHTMLNS        = u"http://www.w3.org/1999/xhtml"
XLINKNS        = u"http://www.w3.org/1999/xlink"
XMLNS          = u"http://www.w3.org/XML/1998/namespace"
XSDNS          = u"http://www.w3.org/2001/XMLSchema"
XSINS          = u"http://www.w3.org/2001/XMLSchema-instance"

nsdict = {
   ANIMNS: u'anim',
   CHARTNS: u'chart',
   CHARTOOONS: u'chartooo',
   CONFIGNS: u'config',
   CSS3TNS: u'css3t',
   DBNS: u'db',
   DCNS: u'dc',
   DOMNS: u'dom',
   DR3DNS: u'dr3d',
   DRAWNS: u'draw',
   FIELDNS: u'field',
   FONS: u'fo',
   FORMNS: u'form',
   FORMXNS: u'formx',
   GRDDLNS: u'grddl',
   KOFFICENS: u'koffice',
   LOEXTNS: u'loext',
   MANIFESTNS: u'manifest',
   MATHNS: u'math',
   METANS: u'meta',
   NUMBERNS: u'number',
   OFFICENS: u'office',
   OFNS: u'of',
   OOONS: u'ooo',
   OOOWNS: u'ooow',
   OOOCNS: u'oooc',
   PRESENTATIONNS: u'presentation',
   RDFANS: u'rdfa',
   RPTNS:  u'rpt',
   SCRIPTNS: u'script',
   SMILNS: u'smil',
   STYLENS: u'style',
   SVGNS: u'svg',
   TABLENS: u'table',
   TABLEOOONS: u'tableooo',
   TEXTNS: u'text',
   XFORMSNS: u'xforms',
   XLINKNS: u'xlink',
   XHTMLNS: u'xhtml',
   XMLNS: u'xml',
   XSDNS: u'xsd',
   XSINS: u'xsi',
}



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2007-2008 Søren Roug, European Environment Agency
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s):
#

# This script is to be embedded in opendocument.py later
# The purpose is to read an ODT/ODP/ODS file and create the datastructure
# in memory. The user should then be able to make operations and then save
# the structure again.

from defusedxml.sax import make_parser
from xml.sax import handler
from xml.sax.xmlreader import InputSource
import xml.sax.saxutils
from odf.element import Element
from odf.namespaces import OFFICENS
try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

#
# Parse the XML files
#
class LoadParser(handler.ContentHandler):
    """ Extract headings from content.xml of an ODT file """
    triggers = (
       (OFFICENS, 'automatic-styles'), (OFFICENS, 'body'),
       (OFFICENS, 'font-face-decls'), (OFFICENS, 'master-styles'),
       (OFFICENS, 'meta'), (OFFICENS, 'scripts'),
       (OFFICENS, 'settings'), (OFFICENS, 'styles') )

    def __init__(self, document):
        self.doc = document
        self.data = []
        self.level = 0
        self.parse = False

    def characters(self, data):
        if self.parse == False:
            return
        self.data.append(data)

    def startElementNS(self, tag, qname, attrs):
        if tag in self.triggers:
            self.parse = True
        if self.doc._parsing != "styles.xml" and tag == (OFFICENS, 'font-face-decls'):
            self.parse = False
        if self.parse == False:
            return

        self.level = self.level + 1
        # Add any accumulated text content
        content = ''.join(self.data)
        if content:
            self.parent.addText(content, check_grammar=False)
            self.data = []
        # Create the element
        attrdict = {}
        for (att,value) in attrs.items():
            attrdict[att] = value
        try:
            e = Element(qname = tag, qattributes=attrdict, check_grammar=False)
            self.curr = e
        except AttributeError as v:
            print ("Error: %s" % v)

        if tag == (OFFICENS, 'automatic-styles'):
            e = self.doc.automaticstyles
        elif tag == (OFFICENS, 'body'):
            e = self.doc.body
        elif tag == (OFFICENS, 'master-styles'):
            e = self.doc.masterstyles
        elif tag == (OFFICENS, 'meta'):
            e = self.doc.meta
        elif tag == (OFFICENS,'scripts'):
            e = self.doc.scripts
        elif tag == (OFFICENS,'settings'):
            e = self.doc.settings
        elif tag == (OFFICENS,'styles'):
            e = self.doc.styles
        elif self.doc._parsing == "styles.xml" and tag == (OFFICENS, 'font-face-decls'):
            e = self.doc.fontfacedecls
        elif hasattr(self,'parent'):
            self.parent.addElement(e, check_grammar=False)
        self.parent = e


    def endElementNS(self, tag, qname):
        if self.parse == False:
            return
        self.level = self.level - 1
        str = ''.join(self.data)
        if str:
            self.curr.addText(str, check_grammar=False)
        self.data = []
        self.curr = self.curr.parentNode
        self.parent = self.curr
        if tag in self.triggers:
            self.parse = False



# -*- coding: utf-8 -*-
# Copyright (C) 2006-2013 Søren Roug, European Environment Agency
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s):
#
import re, sys, os.path
sys.path.append(os.path.dirname(__file__))


from odf.namespaces import TEXTNS
from odf.element import Element
from odf.style import StyleElement

# Autogenerated
def A(**args):
    args.setdefault('type', 'simple')
    return Element(qname = (TEXTNS,'a'), **args)

def AlphabeticalIndex(**args):
    return Element(qname = (TEXTNS,'alphabetical-index'), **args)

def AlphabeticalIndexAutoMarkFile(**args):
    args.setdefault('type', 'simple')
    return Element(qname = (TEXTNS,'alphabetical-index-auto-mark-file'), **args)

def AlphabeticalIndexEntryTemplate(**args):
    return Element(qname = (TEXTNS,'alphabetical-index-entry-template'), **args)

def AlphabeticalIndexMark(**args):
    return Element(qname = (TEXTNS,'alphabetical-index-mark'), **args)

def AlphabeticalIndexMarkEnd(**args):
    return Element(qname = (TEXTNS,'alphabetical-index-mark-end'), **args)

def AlphabeticalIndexMarkStart(**args):
    return Element(qname = (TEXTNS,'alphabetical-index-mark-start'), **args)

def AlphabeticalIndexSource(**args):
    return Element(qname = (TEXTNS,'alphabetical-index-source'), **args)

def AuthorInitials(**args):
    return Element(qname = (TEXTNS,'author-initials'), **args)

def AuthorName(**args):
    return Element(qname = (TEXTNS,'author-name'), **args)

def Bibliography(**args):
    return Element(qname = (TEXTNS,'bibliography'), **args)

def BibliographyConfiguration(**args):
    return Element(qname = (TEXTNS,'bibliography-configuration'), **args)

def BibliographyEntryTemplate(**args):
    return Element(qname = (TEXTNS,'bibliography-entry-template'), **args)

def BibliographyMark(**args):
    return Element(qname = (TEXTNS,'bibliography-mark'), **args)

def BibliographySource(**args):
    return Element(qname = (TEXTNS,'bibliography-source'), **args)

def Bookmark(**args):
    return Element(qname = (TEXTNS,'bookmark'), **args)

def BookmarkEnd(**args):
    return Element(qname = (TEXTNS,'bookmark-end'), **args)

def BookmarkRef(**args):
    return Element(qname = (TEXTNS,'bookmark-ref'), **args)

def BookmarkStart(**args):
    return Element(qname = (TEXTNS,'bookmark-start'), **args)

def Change(**args):
    return Element(qname = (TEXTNS,'change'), **args)

def ChangeEnd(**args):
    return Element(qname = (TEXTNS,'change-end'), **args)

def ChangeStart(**args):
    return Element(qname = (TEXTNS,'change-start'), **args)

def ChangedRegion(**args):
    return Element(qname = (TEXTNS,'changed-region'), **args)

def Chapter(**args):
    return Element(qname = (TEXTNS,'chapter'), **args)

def CharacterCount(**args):
    return Element(qname = (TEXTNS,'character-count'), **args)

def ConditionalText(**args):
    return Element(qname = (TEXTNS,'conditional-text'), **args)

def CreationDate(**args):
    return Element(qname = (TEXTNS,'creation-date'), **args)

def CreationTime(**args):
    return Element(qname = (TEXTNS,'creation-time'), **args)

def Creator(**args):
    return Element(qname = (TEXTNS,'creator'), **args)

def DatabaseDisplay(**args):
    return Element(qname = (TEXTNS,'database-display'), **args)

def DatabaseName(**args):
    return Element(qname = (TEXTNS,'database-name'), **args)

def DatabaseNext(**args):
    return Element(qname = (TEXTNS,'database-next'), **args)

def DatabaseRowNumber(**args):
    return Element(qname = (TEXTNS,'database-row-number'), **args)

def DatabaseRowSelect(**args):
    return Element(qname = (TEXTNS,'database-row-select'), **args)

def Date(**args):
    return Element(qname = (TEXTNS,'date'), **args)

def DdeConnection(**args):
    return Element(qname = (TEXTNS,'dde-connection'), **args)

def DdeConnectionDecl(**args):
    return Element(qname = (TEXTNS,'dde-connection-decl'), **args)

def DdeConnectionDecls(**args):
    return Element(qname = (TEXTNS,'dde-connection-decls'), **args)

def Deletion(**args):
    return Element(qname = (TEXTNS,'deletion'), **args)

def Description(**args):
    return Element(qname = (TEXTNS,'description'), **args)

def EditingCycles(**args):
    return Element(qname = (TEXTNS,'editing-cycles'), **args)

def EditingDuration(**args):
    return Element(qname = (TEXTNS,'editing-duration'), **args)

def ExecuteMacro(**args):
    return Element(qname = (TEXTNS,'execute-macro'), **args)

def Expression(**args):
    return Element(qname = (TEXTNS,'expression'), **args)

def FileName(**args):
    return Element(qname = (TEXTNS,'file-name'), **args)

def FormatChange(**args):
    return Element(qname = (TEXTNS,'format-change'), **args)

def H(**args):
    return Element(qname = (TEXTNS, 'h'), **args)

def HiddenParagraph(**args):
    return Element(qname = (TEXTNS,'hidden-paragraph'), **args)

def HiddenText(**args):
    return Element(qname = (TEXTNS,'hidden-text'), **args)

def IllustrationIndex(**args):
    return Element(qname = (TEXTNS,'illustration-index'), **args)

def IllustrationIndexEntryTemplate(**args):
    return Element(qname = (TEXTNS,'illustration-index-entry-template'), **args)

def IllustrationIndexSource(**args):
    return Element(qname = (TEXTNS,'illustration-index-source'), **args)

def ImageCount(**args):
    return Element(qname = (TEXTNS,'image-count'), **args)

def IndexBody(**args):
    return Element(qname = (TEXTNS,'index-body'), **args)

def IndexEntryBibliography(**args):
    return Element(qname = (TEXTNS,'index-entry-bibliography'), **args)

def IndexEntryChapter(**args):
    return Element(qname = (TEXTNS,'index-entry-chapter'), **args)

def IndexEntryLinkEnd(**args):
    return Element(qname = (TEXTNS,'index-entry-link-end'), **args)

def IndexEntryLinkStart(**args):
    return Element(qname = (TEXTNS,'index-entry-link-start'), **args)

def IndexEntryPageNumber(**args):
    return Element(qname = (TEXTNS,'index-entry-page-number'), **args)

def IndexEntrySpan(**args):
    return Element(qname = (TEXTNS,'index-entry-span'), **args)

def IndexEntryTabStop(**args):
    return Element(qname = (TEXTNS,'index-entry-tab-stop'), **args)

def IndexEntryText(**args):
    return Element(qname = (TEXTNS,'index-entry-text'), **args)

def IndexSourceStyle(**args):
    return Element(qname = (TEXTNS,'index-source-style'), **args)

def IndexSourceStyles(**args):
    return Element(qname = (TEXTNS,'index-source-styles'), **args)

def IndexTitle(**args):
    return Element(qname = (TEXTNS,'index-title'), **args)

def IndexTitleTemplate(**args):
    return Element(qname = (TEXTNS,'index-title-template'), **args)

def InitialCreator(**args):
    return Element(qname = (TEXTNS,'initial-creator'), **args)

def Insertion(**args):
    return Element(qname = (TEXTNS,'insertion'), **args)

def Keywords(**args):
    return Element(qname = (TEXTNS,'keywords'), **args)

def LineBreak(**args):
    return Element(qname = (TEXTNS,'line-break'), **args)

def LinenumberingConfiguration(**args):
    return Element(qname = (TEXTNS,'linenumbering-configuration'), **args)

def LinenumberingSeparator(**args):
    return Element(qname = (TEXTNS,'linenumbering-separator'), **args)

def List(**args):
    return Element(qname = (TEXTNS,'list'), **args)

def ListHeader(**args):
    return Element(qname = (TEXTNS,'list-header'), **args)

def ListItem(**args):
    return Element(qname = (TEXTNS,'list-item'), **args)

def ListLevelStyleBullet(**args):
    return Element(qname = (TEXTNS,'list-level-style-bullet'), **args)

def ListLevelStyleImage(**args):
    return Element(qname = (TEXTNS,'list-level-style-image'), **args)

def ListLevelStyleNumber(**args):
    return Element(qname = (TEXTNS,'list-level-style-number'), **args)

def ListStyle(**args):
    return StyleElement(qname = (TEXTNS,'list-style'), **args)

def Measure(**args):
    return Element(qname = (TEXTNS,'measure'), **args)

def Meta(**args):
    return Element(qname = (TEXTNS,'meta'), **args)

def MetaField(**args):
    return Element(qname = (TEXTNS,'meta-field'), **args)

def ModificationDate(**args):
    return Element(qname = (TEXTNS,'modification-date'), **args)

def ModificationTime(**args):
    return Element(qname = (TEXTNS,'modification-time'), **args)

def Note(**args):
    return Element(qname = (TEXTNS,'note'), **args)

def NoteBody(**args):
    return Element(qname = (TEXTNS,'note-body'), **args)

def NoteCitation(**args):
    return Element(qname = (TEXTNS,'note-citation'), **args)

def NoteContinuationNoticeBackward(**args):
    return Element(qname = (TEXTNS,'note-continuation-notice-backward'), **args)

def NoteContinuationNoticeForward(**args):
    return Element(qname = (TEXTNS,'note-continuation-notice-forward'), **args)

def NoteRef(**args):
    return Element(qname = (TEXTNS,'note-ref'), **args)

def NotesConfiguration(**args):
    return Element(qname = (TEXTNS,'notes-configuration'), **args)

def Number(**args):
    return Element(qname = (TEXTNS,'number'), **args)

def NumberedParagraph(**args):
    return Element(qname = (TEXTNS,'numbered-paragraph'), **args)

def ObjectCount(**args):
    return Element(qname = (TEXTNS,'object-count'), **args)

def ObjectIndex(**args):
    return Element(qname = (TEXTNS,'object-index'), **args)

def ObjectIndexEntryTemplate(**args):
    return Element(qname = (TEXTNS,'object-index-entry-template'), **args)

def ObjectIndexSource(**args):
    return Element(qname = (TEXTNS,'object-index-source'), **args)

def OutlineLevelStyle(**args):
    return Element(qname = (TEXTNS,'outline-level-style'), **args)

def OutlineStyle(**args):
    return Element(qname = (TEXTNS,'outline-style'), **args)

def P(**args):
    return Element(qname = (TEXTNS, 'p'), **args)

def Page(**args):
    return Element(qname = (TEXTNS,'page'), **args)

def PageContinuation(**args):
    return Element(qname = (TEXTNS,'page-continuation'), **args)

def PageCount(**args):
    return Element(qname = (TEXTNS,'page-count'), **args)

def PageNumber(**args):
    return Element(qname = (TEXTNS,'page-number'), **args)

def PageSequence(**args):
    return Element(qname = (TEXTNS,'page-sequence'), **args)

def PageVariableGet(**args):
    return Element(qname = (TEXTNS,'page-variable-get'), **args)

def PageVariableSet(**args):
    return Element(qname = (TEXTNS,'page-variable-set'), **args)

def ParagraphCount(**args):
    return Element(qname = (TEXTNS,'paragraph-count'), **args)

def Placeholder(**args):
    return Element(qname = (TEXTNS,'placeholder'), **args)

def PrintDate(**args):
    return Element(qname = (TEXTNS,'print-date'), **args)

def PrintTime(**args):
    return Element(qname = (TEXTNS,'print-time'), **args)

def PrintedBy(**args):
    return Element(qname = (TEXTNS,'printed-by'), **args)

def ReferenceMark(**args):
    return Element(qname = (TEXTNS,'reference-mark'), **args)

def ReferenceMarkEnd(**args):
    return Element(qname = (TEXTNS,'reference-mark-end'), **args)

def ReferenceMarkStart(**args):
    return Element(qname = (TEXTNS,'reference-mark-start'), **args)

def ReferenceRef(**args):
    return Element(qname = (TEXTNS,'reference-ref'), **args)

def Ruby(**args):
    return Element(qname = (TEXTNS,'ruby'), **args)

def RubyBase(**args):
    return Element(qname = (TEXTNS,'ruby-base'), **args)

def RubyText(**args):
    return Element(qname = (TEXTNS,'ruby-text'), **args)

def S(**args):
    return Element(qname = (TEXTNS,'s'), **args)

def Script(**args):
    return Element(qname = (TEXTNS,'script'), **args)

def Section(**args):
    return Element(qname = (TEXTNS,'section'), **args)

def SectionSource(**args):
    return Element(qname = (TEXTNS,'section-source'), **args)

def SenderCity(**args):
    return Element(qname = (TEXTNS,'sender-city'), **args)

def SenderCompany(**args):
    return Element(qname = (TEXTNS,'sender-company'), **args)

def SenderCountry(**args):
    return Element(qname = (TEXTNS,'sender-country'), **args)

def SenderEmail(**args):
    return Element(qname = (TEXTNS,'sender-email'), **args)

def SenderFax(**args):
    return Element(qname = (TEXTNS,'sender-fax'), **args)

def SenderFirstname(**args):
    return Element(qname = (TEXTNS,'sender-firstname'), **args)

def SenderInitials(**args):
    return Element(qname = (TEXTNS,'sender-initials'), **args)

def SenderLastname(**args):
    return Element(qname = (TEXTNS,'sender-lastname'), **args)

def SenderPhonePrivate(**args):
    return Element(qname = (TEXTNS,'sender-phone-private'), **args)

def SenderPhoneWork(**args):
    return Element(qname = (TEXTNS,'sender-phone-work'), **args)

def SenderPosition(**args):
    return Element(qname = (TEXTNS,'sender-position'), **args)

def SenderPostalCode(**args):
    return Element(qname = (TEXTNS,'sender-postal-code'), **args)

def SenderStateOrProvince(**args):
    return Element(qname = (TEXTNS,'sender-state-or-province'), **args)

def SenderStreet(**args):
    return Element(qname = (TEXTNS,'sender-street'), **args)

def SenderTitle(**args):
    return Element(qname = (TEXTNS,'sender-title'), **args)

def Sequence(**args):
    return Element(qname = (TEXTNS,'sequence'), **args)

def SequenceDecl(**args):
    return Element(qname = (TEXTNS,'sequence-decl'), **args)

def SequenceDecls(**args):
    return Element(qname = (TEXTNS,'sequence-decls'), **args)

def SequenceRef(**args):
    return Element(qname = (TEXTNS,'sequence-ref'), **args)

def SheetName(**args):
    return Element(qname = (TEXTNS,'sheet-name'), **args)

def SoftPageBreak(**args):
    return Element(qname = (TEXTNS,'soft-page-break'), **args)

def SortKey(**args):
    return Element(qname = (TEXTNS,'sort-key'), **args)

def Span(**args):
    return Element(qname = (TEXTNS,'span'), **args)

def Subject(**args):
    return Element(qname = (TEXTNS,'subject'), **args)

def Tab(**args):
    return Element(qname = (TEXTNS,'tab'), **args)

def TableCount(**args):
    return Element(qname = (TEXTNS,'table-count'), **args)

def TableFormula(**args):
    return Element(qname = (TEXTNS,'table-formula'), **args)

def TableIndex(**args):
    return Element(qname = (TEXTNS,'table-index'), **args)

def TableIndexEntryTemplate(**args):
    return Element(qname = (TEXTNS,'table-index-entry-template'), **args)

def TableIndexSource(**args):
    return Element(qname = (TEXTNS,'table-index-source'), **args)

def TableOfContent(**args):
    return Element(qname = (TEXTNS,'table-of-content'), **args)

def TableOfContentEntryTemplate(**args):
    return Element(qname = (TEXTNS,'table-of-content-entry-template'), **args)

def TableOfContentSource(**args):
    return Element(qname = (TEXTNS,'table-of-content-source'), **args)

def TemplateName(**args):
    return Element(qname = (TEXTNS,'template-name'), **args)

def TextInput(**args):
    return Element(qname = (TEXTNS,'text-input'), **args)

def Time(**args):
    return Element(qname = (TEXTNS,'time'), **args)

def Title(**args):
    return Element(qname = (TEXTNS,'title'), **args)

def TocMark(**args):
    return Element(qname = (TEXTNS,'toc-mark'), **args)

def TocMarkEnd(**args):
    return Element(qname = (TEXTNS,'toc-mark-end'), **args)

def TocMarkStart(**args):
    return Element(qname = (TEXTNS,'toc-mark-start'), **args)

def TrackedChanges(**args):
    return Element(qname = (TEXTNS,'tracked-changes'), **args)

def UserDefined(**args):
    return Element(qname = (TEXTNS,'user-defined'), **args)

def UserFieldDecl(**args):
    return Element(qname = (TEXTNS,'user-field-decl'), **args)

def UserFieldDecls(**args):
    return Element(qname = (TEXTNS,'user-field-decls'), **args)

def UserFieldGet(**args):
    return Element(qname = (TEXTNS,'user-field-get'), **args)

def UserFieldInput(**args):
    return Element(qname = (TEXTNS,'user-field-input'), **args)

def UserIndex(**args):
    return Element(qname = (TEXTNS,'user-index'), **args)

def UserIndexEntryTemplate(**args):
    return Element(qname = (TEXTNS,'user-index-entry-template'), **args)

def UserIndexMark(**args):
    return Element(qname = (TEXTNS,'user-index-mark'), **args)

def UserIndexMarkEnd(**args):
    return Element(qname = (TEXTNS,'user-index-mark-end'), **args)

def UserIndexMarkStart(**args):
    return Element(qname = (TEXTNS,'user-index-mark-start'), **args)

def UserIndexSource(**args):
    return Element(qname = (TEXTNS,'user-index-source'), **args)

def VariableDecl(**args):
    return Element(qname = (TEXTNS,'variable-decl'), **args)

def VariableDecls(**args):
    return Element(qname = (TEXTNS,'variable-decls'), **args)

def VariableGet(**args):
    return Element(qname = (TEXTNS,'variable-get'), **args)

def VariableInput(**args):
    return Element(qname = (TEXTNS,'variable-input'), **args)

def VariableSet(**args):
    return Element(qname = (TEXTNS,'variable-set'), **args)

def WordCount(**args):
    return Element(qname = (TEXTNS,'word-count'), **args)




# -*- coding: utf-8 -*-
# Copyright (C) 2006-2008 Søren Roug, European Environment Agency
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# See http://trac.edgewall.org/wiki/WikiFormatting
#
# Contributor(s):
#

import sys, zipfile, xml.dom.minidom
from odf.namespaces import nsdict
from odf.elementtypes import *

IGNORED_TAGS = [
    'draw:a'
    'draw:g',
    'draw:line',
    'draw:object-ole',
    'office:annotation',
    'presentation:notes',
    'svg:desc',
] + [ nsdict[item[0]]+":"+item[1] for item in empty_elements]

INLINE_TAGS = [ nsdict[item[0]]+":"+item[1] for item in inline_elements]


class TextProps:
    """ Holds properties for a text style. """

    def __init__(self):

        self.italic = False
        self.bold = False
        self.fixed = False
        self.underlined = False
        self.strikethrough = False
        self.superscript = False
        self.subscript = False

    def setItalic(self, value):
        if value == "italic":
            self.italic = True
        elif value == "normal":
            self.italic = False

    def setBold(self, value):
        if value == "bold":
            self.bold = True
        elif value == "normal":
            self.bold = False

    def setFixed(self, value):
        self.fixed = value

    def setUnderlined(self, value):
        if value and value != "none":
            self.underlined = True

    def setStrikethrough(self, value):
        if value and value != "none":
            self.strikethrough = True

    def setPosition(self, value):
        if value is None or value == '':
            return
        posisize = value.split(' ')
        textpos = posisize[0]
        if textpos.find('%') == -1:
            if textpos == "sub":
                self.superscript = False
                self.subscript = True
            elif textpos == "super":
                self.superscript = True
                self.subscript = False
        else:
            itextpos = int(textpos[:textpos.find('%')])
            if itextpos > 10:
                self.superscript = False
                self.subscript = True
            elif itextpos < -10:
                self.superscript = True
                self.subscript = False

    def __str__(self):

        return "[italic=%s, bold=i%s, fixed=%s]" % (str(self.italic),
                                          str(self.bold),
                                          str(self.fixed))

class ParagraphProps:
    """ Holds properties of a paragraph style. """

    def __init__(self):

        self.blockquote = False
        self.headingLevel = 0
        self.code = False
        self.title = False
        self.indented = 0

    def setIndented(self, value):
        self.indented = value

    def setHeading(self, level):
        self.headingLevel = level

    def setTitle(self, value):
        self.title = value

    def setCode(self, value):
        self.code = value


    def __str__(self):

        return "[bq=%s, h=%d, code=%s]" % (str(self.blockquote),
                                           self.headingLevel,
                                           str(self.code))


class ListProperties:
    """ Holds properties for a list style. """

    def __init__(self):
        self.ordered = False

    def setOrdered(self, value):
        self.ordered = value



class ODF2MoinMoin(object):


    def __init__(self, filepath):
        self.footnotes = []
        self.footnoteCounter = 0
        self.textStyles = {"Standard": TextProps()}
        self.paragraphStyles = {"Standard": ParagraphProps()}
        self.listStyles = {}
        self.fixedFonts = []
        self.hasTitle = 0
        self.lastsegment = None

        # Tags
        self.elements = {
         'draw:page': self.textToString,
         'draw:frame': self.textToString,
         'draw:image': self.draw_image,
         'draw:text-box': self.textToString,
         'text:a': self.text_a,
         'text:note': self.text_note,
        }
        for tag in IGNORED_TAGS:
            self.elements[tag] = self.do_nothing

        for tag in INLINE_TAGS:
            self.elements[tag] = self.inline_markup
        self.elements['text:line-break'] = self.text_line_break
        self.elements['text:s'] = self.text_s
        self.elements['text:tab'] = self.text_tab

        self.load(filepath)

    def processFontDeclarations(self, fontDecl):
        """ Extracts necessary font information from a font-declaration
            element.
            """
        for fontFace in fontDecl.getElementsByTagName("style:font-face"):
            if fontFace.getAttribute("style:font-pitch") == "fixed":
                self.fixedFonts.append(fontFace.getAttribute("style:name"))



    def extractTextProperties(self, style, parent=None):
        """ Extracts text properties from a style element. """

        textProps = TextProps()

        if parent:
            parentProp = self.textStyles.get(parent, None)
            if parentProp:
                textProp = parentProp

        textPropEl = style.getElementsByTagName("style:text-properties")
        if not textPropEl: return textProps

        textPropEl = textPropEl[0]

        textProps.setItalic(textPropEl.getAttribute("fo:font-style"))
        textProps.setBold(textPropEl.getAttribute("fo:font-weight"))
        textProps.setUnderlined(textPropEl.getAttribute("style:text-underline-style"))
        textProps.setStrikethrough(textPropEl.getAttribute("style:text-line-through-style"))
        textProps.setPosition(textPropEl.getAttribute("style:text-position"))

        if textPropEl.getAttribute("style:font-name") in self.fixedFonts:
            textProps.setFixed(True)

        return textProps

    def extractParagraphProperties(self, style, parent=None):
        """ Extracts paragraph properties from a style element. """

        paraProps = ParagraphProps()

        name = style.getAttribute("style:name")

        if name.startswith("Heading_20_"):
            level = name[11:]
            try:
                level = int(level)
                paraProps.setHeading(level)
            except:
                level = 0

        if name == "Title":
            paraProps.setTitle(True)

        paraPropEl = style.getElementsByTagName("style:paragraph-properties")
        if paraPropEl:
            paraPropEl = paraPropEl[0]
            leftMargin = paraPropEl.getAttribute("fo:margin-left")
            if leftMargin:
                try:
                    leftMargin = float(leftMargin[:-2])
                    if leftMargin > 0.01:
                        paraProps.setIndented(True)
                except:
                    pass

        textProps = self.extractTextProperties(style)
        if textProps.fixed:
            paraProps.setCode(True)

        return paraProps


    def processStyles(self, styleElements):
        """ Runs through "style" elements extracting necessary information.
        """

        for style in styleElements:

            name = style.getAttribute("style:name")

            if name == "Standard": continue

            family = style.getAttribute("style:family")
            parent = style.getAttribute("style:parent-style-name")

            if family == "text":
                self.textStyles[name] = self.extractTextProperties(style, parent)

            elif family == "paragraph":
                self.paragraphStyles[name] = \
                                 self.extractParagraphProperties(style, parent)
                self.textStyles[name] = self.extractTextProperties(style, parent)

    def processListStyles(self, listStyleElements):

        for style in listStyleElements:
            name = style.getAttribute("style:name")

            prop = ListProperties()
            if style.hasChildNodes():
                subitems = [el for el in style.childNodes
                     if el.nodeType == xml.dom.Node.ELEMENT_NODE
                     and el.tagName == "text:list-level-style-number"]
                if len(subitems) > 0:
                    prop.setOrdered(True)

            self.listStyles[name] = prop


    def load(self, filepath):
        """ Loads an ODT file. """

        zip = zipfile.ZipFile(filepath)

        styles_doc = xml.dom.minidom.parseString(zip.read("styles.xml"))
        fontfacedecls = styles_doc.getElementsByTagName("office:font-face-decls")
        if fontfacedecls:
            self.processFontDeclarations(fontfacedecls[0])
        self.processStyles(styles_doc.getElementsByTagName("style:style"))
        self.processListStyles(styles_doc.getElementsByTagName("text:list-style"))

        self.content = xml.dom.minidom.parseString(zip.read("content.xml"))
        fontfacedecls = self.content.getElementsByTagName("office:font-face-decls")
        if fontfacedecls:
            self.processFontDeclarations(fontfacedecls[0])

        self.processStyles(self.content.getElementsByTagName("style:style"))
        self.processListStyles(self.content.getElementsByTagName("text:list-style"))

    def compressCodeBlocks(self, text):
        """ Removes extra blank lines from code blocks. """

        return text
        lines = text.split("\n")
        buffer = []
        numLines = len(lines)
        for i in range(numLines):

            if (lines[i].strip() or i == numLines-1  or i == 0 or
                not ( lines[i-1].startswith("    ")
                      and lines[i+1].startswith("    ") ) ):
                buffer.append("\n" + lines[i])

        return ''.join(buffer)

#-----------------------------------
    def do_nothing(self, node):
        return ''

    def draw_image(self, node):
        """
        """

        link = node.getAttribute("xlink:href")
        if link and link[:2] == './': # Indicates a sub-object, which isn't supported
            return "%s\n" % link
        if link and link[:9] == 'Pictures/':
            link = link[9:]
        return "[[Image(%s)]]\n" % link

    def text_a(self, node):
        text = self.textToString(node)
        link = node.getAttribute("xlink:href")
        if link.strip() == text.strip():
            return "[%s] " % link.strip()
        else:
            return "[%s %s] " % (link.strip(), text.strip())


    def text_line_break(self, node):
        return "[[BR]]"

    def text_note(self, node):
        cite = (node.getElementsByTagName("text:note-citation")[0]
                    .childNodes[0].nodeValue)
        body = (node.getElementsByTagName("text:note-body")[0]
                    .childNodes[0])
        self.footnotes.append((cite, self.textToString(body)))
        return "^%s^" % cite

    def text_s(self, node):
        try:
            num = int(node.getAttribute("text:c"))
            return " "*num
        except:
            return " "

    def text_tab(self, node):
        return "    "

    def inline_markup(self, node):
        text = self.textToString(node)

        if not text.strip():
            return ''  # don't apply styles to white space

        styleName = node.getAttribute("text:style-name")
        style = self.textStyles.get(styleName, TextProps())

        if style.fixed:
            return "`" + text + "`"

        mark = []
        if style:
            if style.italic:
                mark.append("''")
            if style.bold:
                mark.append("'''")
            if style.underlined:
                mark.append("__")
            if style.strikethrough:
                mark.append("~~")
            if style.superscript:
                mark.append("^")
            if style.subscript:
                mark.append(",,")
        revmark = mark[:]
        revmark.reverse()
        return "%s%s%s" % (''.join(mark), text, ''.join(revmark))

#-----------------------------------
    def listToString(self, listElement, indent = 0):

        self.lastsegment = listElement.tagName
        buffer = []

        styleName = listElement.getAttribute("text:style-name")
        props = self.listStyles.get(styleName, ListProperties())

        i = 0
        for item in listElement.childNodes:
            buffer.append(" "*indent)
            i += 1
            if props.ordered:
                number = str(i)
                number = " " + number + ". "
                buffer.append(" 1. ")
            else:
                buffer.append(" * ")
            subitems = [el for el in item.childNodes
                          if el.tagName in ["text:p", "text:h", "text:list"]]
            for subitem in subitems:
                if subitem.tagName == "text:list":
                    buffer.append("\n")
                    buffer.append(self.listToString(subitem, indent+3))
                else:
                    buffer.append(self.paragraphToString(subitem, indent+3))
                self.lastsegment = subitem.tagName
            self.lastsegment = item.tagName
            buffer.append("\n")

        return ''.join(buffer)

    def tableToString(self, tableElement):
        """ MoinMoin uses || to delimit table cells
        """

        self.lastsegment = tableElement.tagName
        buffer = []

        for item in tableElement.childNodes:
            self.lastsegment = item.tagName
            if item.tagName == "table:table-header-rows":
                buffer.append(self.tableToString(item))
            if item.tagName == "table:table-row":
                buffer.append("\n||")
                for cell in item.childNodes:
                    buffer.append(self.inline_markup(cell))
                    buffer.append("||")
                    self.lastsegment = cell.tagName
        return ''.join(buffer)


    def toString(self):
        """ Converts the document to a string.
            FIXME: Result from second call differs from first call
        """
        body = self.content.getElementsByTagName("office:body")[0]
        text = body.childNodes[0]

        buffer = []

        paragraphs = [el for el in text.childNodes
                      if el.tagName in ["draw:page", "text:p", "text:h","text:section",
                                        "text:list", "table:table"]]

        for paragraph in paragraphs:
            if paragraph.tagName == "text:list":
                text = self.listToString(paragraph)
            elif paragraph.tagName == "text:section":
                text = self.textToString(paragraph)
            elif paragraph.tagName == "table:table":
                text = self.tableToString(paragraph)
            else:
                text = self.paragraphToString(paragraph)
            if text:
                buffer.append(text)

        if self.footnotes:

            buffer.append("----")
            for cite, body in self.footnotes:
                buffer.append("%s: %s" % (cite, body))


        buffer.append("")
        return self.compressCodeBlocks('\n'.join(buffer))


    def textToString(self, element):

        buffer = []

        for node in element.childNodes:

            if node.nodeType == xml.dom.Node.TEXT_NODE:
                buffer.append(node.nodeValue)

            elif node.nodeType == xml.dom.Node.ELEMENT_NODE:
                tag = node.tagName

                if tag in ("draw:text-box", "draw:frame"):
                    buffer.append(self.textToString(node))

                elif tag in ("text:p", "text:h"):
                    text = self.paragraphToString(node)
                    if text:
                        buffer.append(text)
                elif tag == "text:list":
                    buffer.append(self.listToString(node))
                else:
                    method = self.elements.get(tag)
                    if method:
                        buffer.append(method(node))
                    else:
                        buffer.append(" {" + tag + "} ")

        return ''.join(buffer)

    def paragraphToString(self, paragraph, indent = 0):

        dummyParaProps = ParagraphProps()

        style_name = paragraph.getAttribute("text:style-name")
        paraProps = self.paragraphStyles.get(style_name, dummyParaProps)
        text = self.inline_markup(paragraph)

        if paraProps and not paraProps.code:
            text = text.strip()

        if paragraph.tagName == "text:p" and self.lastsegment == "text:p":
            text = "\n" + text

        self.lastsegment = paragraph.tagName

        if paraProps.title:
            self.hasTitle = 1
            return "= " + text + " =\n"

        outlinelevel = paragraph.getAttribute("text:outline-level")
        if outlinelevel:

            level = int(outlinelevel)
            if self.hasTitle: level += 1

            if level >= 1:
                return "=" * level + " " + text + " " + "=" * level + "\n"

        elif paraProps.code:
            return "{{{\n" + text + "\n}}}\n"

        if paraProps.indented:
            return self.wrapParagraph(text, indent = indent, blockquote = True)

        else:
            return self.wrapParagraph(text, indent = indent)


    def wrapParagraph(self, text, indent = 0, blockquote=False):

        counter = 0
        buffer = []
        LIMIT = 50

        if blockquote:
            buffer.append("  ")

        return ''.join(buffer) + text
        # Unused from here
        for token in text.split():

            if counter > LIMIT - indent:
                buffer.append("\n" + " "*indent)
                if blockquote:
                    buffer.append("  ")
                counter = 0

            buffer.append(token + " ")
            counter += len(token)

        return ''.join(buffer)



# -*- coding: utf-8 -*-
#
#   Create and extract text from ODF, handling whitespace correctly.
#   Copyright (C) 2008 J. David Eisenberg
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


"""
Class for handling whitespace properly in OpenDocument.

While it is possible to use getTextContent() and setTextContent()
to extract or create ODF content, these won't extract or create
the appropriate <text:s>, <text:tab>, or <text:line-break>
elements.  This module takes care of that problem.
"""

from odf.element import Node
import odf.opendocument
from odf.text import S,LineBreak,Tab

class WhitespaceText(object):

    def __init__(self):
        self.textBuffer = []
        self.spaceCount = 0

    def addTextToElement(self, odfElement, s):
        """ Process an input string, inserting
            <text:tab> elements for '\t',
            <text:line-break> elements for '\n', and
            <text:s> elements for runs of more than one blank.
            These will be added to the given element.
        """
        i = 0
        ch = ' '

        # When we encounter a tab or newline, we can immediately
        # dump any accumulated text and then emit the appropriate
        # ODF element.
        #
        # When we encounter a space, we add it to the text buffer,
        # and then collect more spaces.  If there are more spaces
        # after the first one, we dump the text buffer and then
        # then emit the appropriate <text:s> element.

        while i < len(s):
            ch = s[i]
            if ch == '\t':
                self._emitTextBuffer(odfElement)
                odfElement.addElement(Tab())
                i += 1
            elif ch == '\n':
                self._emitTextBuffer(odfElement);
                odfElement.addElement(LineBreak())
                i += 1
            elif ch == ' ':
                self.textBuffer.append(' ')
                i += 1
                self.spaceCount = 0
                while i < len(s) and (s[i] == ' '):
                    self.spaceCount += 1
                    i += 1
                if self.spaceCount > 0:
                    self._emitTextBuffer(odfElement)
                    self._emitSpaces(odfElement)
            else:
                self.textBuffer.append(ch)
                i += 1

        self._emitTextBuffer(odfElement)

    def _emitTextBuffer(self, odfElement):
        """ Creates a Text Node whose contents are the current textBuffer.
            Side effect: clears the text buffer.
        """
        if len(self.textBuffer) > 0:
            odfElement.addText(''.join(self.textBuffer))
        self.textBuffer = []


    def _emitSpaces(self, odfElement):
        """ Creates a <text:s> element for the current spaceCount.
            Side effect: sets spaceCount back to zero
        """
        if self.spaceCount > 0:
            spaceElement = S(c=self.spaceCount)
            odfElement.addElement(spaceElement)
        self.spaceCount = 0

def addTextToElement(odfElement, s):
    wst = WhitespaceText()
    wst.addTextToElement(odfElement, s)

def extractText(odfElement):
    """ Extract text content from an Element, with whitespace represented
        properly. Returns the text, with tabs, spaces, and newlines
        correctly evaluated. This method recursively descends through the
        children of the given element, accumulating text and "unwrapping"
        <text:s>, <text:tab>, and <text:line-break> elements along the way.
    """
    result = [];

    if len(odfElement.childNodes) != 0:
        for child in odfElement.childNodes:
            if child.nodeType == Node.TEXT_NODE:
                result.append(child.data)
            elif child.nodeType == Node.ELEMENT_NODE:
                subElement = child
                tagName = subElement.qname;
                if tagName == (u"urn:oasis:names:tc:opendocument:xmlns:text:1.0", u"line-break"):
                    result.append("\n")
                elif tagName == (u"urn:oasis:names:tc:opendocument:xmlns:text:1.0", u"tab"):
                    result.append("\t")
                elif tagName == (u"urn:oasis:names:tc:opendocument:xmlns:text:1.0", u"s"):
                    c = subElement.getAttribute('c')
                    if c:
                        spaceCount =  int(c)
                    else:
                        spaceCount = 1

                    result.append(" " * spaceCount)
                else:
                    result.append(extractText(subElement))
    return ''.join(result)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2006-2009 Søren Roug, European Environment Agency
#
# This is free software.  You may redistribute it under the terms
# of the Apache license and the GNU General Public License Version
# 2 or at your option any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s): Michael Howitz, gocept gmbh & co. kg
#
# $Id: userfield.py 447 2008-07-10 20:01:30Z roug $

"""Class to show and manipulate user fields in odf documents."""

import sys
import zipfile

from odf.text import UserFieldDecl
from odf.namespaces import OFFICENS
from odf.opendocument import load
import io, sys

if sys.version_info[0]==3:
    unicode=str

OUTENCODING = "utf-8"


# OpenDocument v.1.0 section 6.7.1
VALUE_TYPES = {
    u'float': (OFFICENS, u'value'),
    u'percentage': (OFFICENS, u'value'),
    u'currency': (OFFICENS, u'value'),
    u'date': (OFFICENS, u'date-value'),
    u'time': (OFFICENS, u'time-value'),
    u'boolean': (OFFICENS, u'boolean-value'),
    u'string': (OFFICENS, u'string-value'),
    }


class UserFields(object):
    """List, view and manipulate user fields."""

    # these attributes can be a filename or a file like object
    src_file = None
    dest_file = None

    def __init__(self, src=None, dest=None):
        """Constructor

        @param src open file in binary mode: source document,
        or filename as a unicode string, or None for stdin.
        @param dest opendile in binary mode: destination document,
        or filename as a unicode string, or None for stdout.
        """
        assert(src==None or 'rb' in repr(src) or 'BufferedReader' in repr(src) or 'BytesIO' in repr(src) or type(src)==type(u""))
        assert(dest==None or 'wb' in repr(dest) or 'BufferedWriter' in repr(dest) or 'BytesIO' in repr(dest) or type(dest)==type(u""))
        self.src_file = src
        self.dest_file = dest
        self.document = None

    def loaddoc(self):
        if (sys.version_info[0]==3 and (isinstance(self.src_file, str) or (isinstance(self.src_file, io.IOBase)))) or (sys.version_info[0]==2 and isinstance(self.src_file, basestring)):
            # src_file is a filename, check if it is a zip-file
            if not zipfile.is_zipfile(self.src_file):
                raise TypeError(u"%s is no odt file." % self.src_file)
        elif self.src_file is None:
            # use stdin if no file given
            self.src_file = sys.stdin

        self.document = load(self.src_file)

    def savedoc(self):
        # write output
        if self.dest_file is None:
            # use stdout if no filename given
            self.document.save(u'-')
        else:
            self.document.save(self.dest_file)

    def list_fields(self):
        """List (extract) all known user-fields.

        @return list of user-field names as unicode strings.
        """
        return [x[0] for x in self.list_fields_and_values()]

    def list_fields_and_values(self, field_names=None):
        """List (extract) user-fields with type and value.

        @param field_names list of field names as unicode strings
        to show, or None for all.

        @return list of tuples (<field name>, <field type>, <value>)
        as type (unicode string, stringified type, unicode string).

        """
        self.loaddoc()
        found_fields = []
        all_fields = self.document.getElementsByType(UserFieldDecl)
        for f in all_fields:
            value_type = f.getAttribute(u'valuetype')
            if value_type == u'string':
                value = f.getAttribute(u'stringvalue')
            else:
                value = f.getAttribute(u'value')
            field_name = f.getAttribute(u'name')

            if field_names is None or field_name in field_names:
                found_fields.append((field_name,
                                     value_type,
                                     value))
        return found_fields

    def list_values(self, field_names):
        """Extract the contents of given field names from the file.

        @param field_names list of field names as unicode strings

        @return list of field values as unicode strings.

        """
        return [x[2] for x in self.list_fields_and_values(field_names)]

    def get(self, field_name):
        """Extract the contents of this field from the file.
        @param field_name unicode string: name of a field
        @return field value as a unicode string or None if field does not exist.

        """
        assert(type(field_name)==type(u""))
        values = self.list_values([field_name])
        if not values:
            return None
        return values[0]

    def get_type_and_value(self, field_name):
        """Extract the type and contents of this field from the file.
        @param field_name unicode string: name of a field
        @return tuple (<type>, <field-value>) as a pair of unicode strings
        or None if field does not exist.

        """
        assert(type(field_name)==type(u""))
        fields = self.list_fields_and_values([field_name])
        if not fields:
            return None
        field_name, value_type, value = fields[0]
        return value_type, value

    def update(self, data):
        """Set the value of user fields. The field types will be the same.

        data ... dict, with field name as key, field value as value

        Returns None

        """
        self.loaddoc()
        all_fields = self.document.getElementsByType(UserFieldDecl)
        for f in all_fields:
            field_name = f.getAttribute(u'name')
            if field_name in data:
                value_type = f.getAttribute(u'valuetype')
                value = data.get(field_name)
                if value_type == u'string':
                    f.setAttribute(u'stringvalue', value)
                else:
                    f.setAttribute(u'value', value)
        self.savedoc()




# -*- coding: utf-8 -*-
# Copyright (C) 2007 Søren Roug, European Environment Agency
#
# This is free software.  You may redistribute it under the terms
# of the Apache license and the GNU General Public License Version
# 2 or at your option any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s):
#




# -*- coding: utf-8 -*-
# Copyright (C) 2006-2007 Søren Roug, European Environment Agency
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s):
#
import sys, os.path
sys.path.append(os.path.dirname(__file__))

from odf.namespaces import DR3DNS
from odf.element import Element
from odf.draw import StyleRefElement

# Autogenerated
def Cube(**args):
    return StyleRefElement(qname = (DR3DNS,'cube'), **args)

def Extrude(**args):
    return StyleRefElement(qname = (DR3DNS,'extrude'), **args)

def Light(Element):
    return StyleRefElement(qname = (DR3DNS,'light'), **args)

def Rotate(**args):
    return StyleRefElement(qname = (DR3DNS,'rotate'), **args)

def Scene(**args):
    return StyleRefElement(qname = (DR3DNS,'scene'), **args)

def Sphere(**args):
    return StyleRefElement(qname = (DR3DNS,'sphere'), **args)




# -*- coding: utf-8 -*-
# Copyright (C) 2006-2013 Søren Roug, European Environment Agency
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s):
#

from odf.namespaces import TABLENS
from odf.element import Element


# Autogenerated
def Background(**args):
    return Element(qname = (TABLENS,'background'), **args)

def Body(**args):
    return Element(qname = (TABLENS,'body'), **args)

def CalculationSettings(**args):
    return Element(qname = (TABLENS,'calculation-settings'), **args)

def CellAddress(**args):
    return Element(qname = (TABLENS,'cell-address'), **args)

def CellContentChange(**args):
    return Element(qname = (TABLENS,'cell-content-change'), **args)

def CellContentDeletion(**args):
    return Element(qname = (TABLENS,'cell-content-deletion'), **args)

def CellRangeSource(**args):
    args.setdefault('type', 'simple')
    return Element(qname = (TABLENS,'cell-range-source'), **args)

def ChangeDeletion(**args):
    return Element(qname = (TABLENS,'change-deletion'), **args)

def ChangeTrackTableCell(**args):
    return Element(qname = (TABLENS,'change-track-table-cell'), **args)

def Consolidation(**args):
    return Element(qname = (TABLENS,'consolidation'), **args)

def ContentValidation(**args):
    return Element(qname = (TABLENS,'content-validation'), **args)

def ContentValidations(**args):
    return Element(qname = (TABLENS,'content-validations'), **args)

def CoveredTableCell(**args):
    return Element(qname = (TABLENS,'covered-table-cell'), **args)

def CutOffs(**args):
    return Element(qname = (TABLENS,'cut-offs'), **args)

def DataPilotDisplayInfo(**args):
    return Element(qname = (TABLENS,'data-pilot-display-info'), **args)

def DataPilotField(**args):
    return Element(qname = (TABLENS,'data-pilot-field'), **args)

def DataPilotFieldReference(**args):
    return Element(qname = (TABLENS,'data-pilot-field-reference'), **args)

def DataPilotGroup(**args):
    return Element(qname = (TABLENS,'data-pilot-group'), **args)

def DataPilotGroupMember(**args):
    return Element(qname = (TABLENS,'data-pilot-group-member'), **args)

def DataPilotGroups(**args):
    return Element(qname = (TABLENS,'data-pilot-groups'), **args)

def DataPilotLayoutInfo(**args):
    return Element(qname = (TABLENS,'data-pilot-layout-info'), **args)

def DataPilotLevel(**args):
    return Element(qname = (TABLENS,'data-pilot-level'), **args)

def DataPilotMember(**args):
    return Element(qname = (TABLENS,'data-pilot-member'), **args)

def DataPilotMembers(**args):
    return Element(qname = (TABLENS,'data-pilot-members'), **args)

def DataPilotSortInfo(**args):
    return Element(qname = (TABLENS,'data-pilot-sort-info'), **args)

def DataPilotSubtotal(**args):
    return Element(qname = (TABLENS,'data-pilot-subtotal'), **args)

def DataPilotSubtotals(**args):
    return Element(qname = (TABLENS,'data-pilot-subtotals'), **args)

def DataPilotTable(**args):
    return Element(qname = (TABLENS,'data-pilot-table'), **args)

def DataPilotTables(**args):
    return Element(qname = (TABLENS,'data-pilot-tables'), **args)

def DatabaseRange(**args):
    return Element(qname = (TABLENS,'database-range'), **args)

def DatabaseRanges(**args):
    return Element(qname = (TABLENS,'database-ranges'), **args)

def DatabaseSourceQuery(**args):
    return Element(qname = (TABLENS,'database-source-query'), **args)

def DatabaseSourceSql(**args):
    return Element(qname = (TABLENS,'database-source-sql'), **args)

def DatabaseSourceTable(**args):
    return Element(qname = (TABLENS,'database-source-table'), **args)

def DdeLink(**args):
    return Element(qname = (TABLENS,'dde-link'), **args)

def DdeLinks(**args):
    return Element(qname = (TABLENS,'dde-links'), **args)

def Deletion(**args):
    return Element(qname = (TABLENS,'deletion'), **args)

def Deletions(**args):
    return Element(qname = (TABLENS,'deletions'), **args)

def Dependencies(**args):
    return Element(qname = (TABLENS,'dependencies'), **args)

def Dependency(**args):
    return Element(qname = (TABLENS,'dependency'), **args)

def Desc(**args):
    return Element(qname = (TABLENS,'desc'), **args)

def Detective(**args):
    return Element(qname = (TABLENS,'detective'), **args)

def ErrorMacro(**args):
    return Element(qname = (TABLENS,'error-macro'), **args)

def ErrorMessage(**args):
    return Element(qname = (TABLENS,'error-message'), **args)

def EvenColumns(**args):
    return Element(qname = (TABLENS,'even-columns'), **args)

def EvenRows(**args):
    return Element(qname = (TABLENS,'even-rows'), **args)

def Filter(**args):
    return Element(qname = (TABLENS,'filter'), **args)

def FilterAnd(**args):
    return Element(qname = (TABLENS,'filter-and'), **args)

def FilterCondition(**args):
    return Element(qname = (TABLENS,'filter-condition'), **args)

def FilterOr(**args):
    return Element(qname = (TABLENS,'filter-or'), **args)

def FilterSetItem(**args):
    return Element(qname = (TABLENS,'filter-set-item'), **args)

def FirstColumn(**args):
    return Element(qname = (TABLENS,'first-column'), **args)

def FirstRow(**args):
    return Element(qname = (TABLENS,'first-row'), **args)

def HelpMessage(**args):
    return Element(qname = (TABLENS,'help-message'), **args)

def HighlightedRange(**args):
    return Element(qname = (TABLENS,'highlighted-range'), **args)

def Insertion(**args):
    return Element(qname = (TABLENS,'insertion'), **args)

def InsertionCutOff(**args):
    return Element(qname = (TABLENS,'insertion-cut-off'), **args)

def Iteration(**args):
    return Element(qname = (TABLENS,'iteration'), **args)

def LabelRange(**args):
    return Element(qname = (TABLENS,'label-range'), **args)

def LabelRanges(**args):
    return Element(qname = (TABLENS,'label-ranges'), **args)

def LastColumn(**args):
    return Element(qname = (TABLENS,'last-column'), **args)

def LastRow(**args):
    return Element(qname = (TABLENS,'last-row'), **args)

def Movement(**args):
    return Element(qname = (TABLENS,'movement'), **args)

def MovementCutOff(**args):
    return Element(qname = (TABLENS,'movement-cut-off'), **args)

def NamedExpression(**args):
    return Element(qname = (TABLENS,'named-expression'), **args)

def NamedExpressions(**args):
    return Element(qname = (TABLENS,'named-expressions'), **args)

def NamedRange(**args):
    return Element(qname = (TABLENS,'named-range'), **args)

def NullDate(**args):
    return Element(qname = (TABLENS,'null-date'), **args)

def OddColumns(**args):
    return Element(qname = (TABLENS,'odd-columns'), **args)

def OddRows(**args):
    return Element(qname = (TABLENS,'odd-rows'), **args)

def Operation(**args):
    return Element(qname = (TABLENS,'operation'), **args)

def Previous(**args):
    return Element(qname = (TABLENS,'previous'), **args)

def Scenario(**args):
    return Element(qname = (TABLENS,'scenario'), **args)

def Shapes(**args):
    return Element(qname = (TABLENS,'shapes'), **args)

def Sort(**args):
    return Element(qname = (TABLENS,'sort'), **args)

def SortBy(**args):
    return Element(qname = (TABLENS,'sort-by'), **args)

def SortGroups(**args):
    return Element(qname = (TABLENS,'sort-groups'), **args)

def SourceCellRange(**args):
    return Element(qname = (TABLENS,'source-cell-range'), **args)

def SourceRangeAddress(**args):
    return Element(qname = (TABLENS,'source-range-address'), **args)

def SourceService(**args):
    return Element(qname = (TABLENS,'source-service'), **args)

def SubtotalField(**args):
    return Element(qname = (TABLENS,'subtotal-field'), **args)

def SubtotalRule(**args):
    return Element(qname = (TABLENS,'subtotal-rule'), **args)

def SubtotalRules(**args):
    return Element(qname = (TABLENS,'subtotal-rules'), **args)

def Table(**args):
    return Element(qname = (TABLENS,'table'), **args)

def TableCell(**args):
    return Element(qname = (TABLENS,'table-cell'), **args)

def TableColumn(**args):
    return Element(qname = (TABLENS,'table-column'), **args)

def TableColumnGroup(**args):
    return Element(qname = (TABLENS,'table-column-group'), **args)

def TableColumns(**args):
    return Element(qname = (TABLENS,'table-columns'), **args)

def TableHeaderColumns(**args):
    return Element(qname = (TABLENS,'table-header-columns'), **args)

def TableHeaderRows(**args):
    return Element(qname = (TABLENS,'table-header-rows'), **args)

def TableRow(**args):
    return Element(qname = (TABLENS,'table-row'), **args)

def TableRowGroup(**args):
    return Element(qname = (TABLENS,'table-row-group'), **args)

def TableRows(**args):
    return Element(qname = (TABLENS,'table-rows'), **args)

def TableSource(**args):
    args.setdefault('type', 'simple')
    return Element(qname = (TABLENS,'table-source'), **args)

def TableTemplate(**args):
    return Element(qname = (TABLENS,'table-template'), **args)

def TargetRangeAddress(**args):
    return Element(qname = (TABLENS,'target-range-address'), **args)

def Title(**args):
    return Element(qname = (TABLENS,'title'), **args)

def TrackedChanges(**args):
    return Element(qname = (TABLENS,'tracked-changes'), **args)

